from __future__ import annotations

import dataclasses
import logging
from enum import Enum
from typing import Dict, FrozenSet, List, Optional, Sequence, Set, Tuple

from dbt_semantic_interfaces.enum_extension import assert_values_exhausted
from dbt_semantic_interfaces.references import EntityReference, SemanticModelReference, TimeDimensionReference
from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.semantics.linkable_element import LinkableElementType
from metricflow_semantics.model.semantics.semantic_model_join_evaluator import MAX_JOIN_HOPS
from metricflow_semantics.model.semantics.semantic_model_lookup import SemanticModelLookup
from metricflow_semantics.specs.entity_spec import LinklessEntitySpec
from metricflow_semantics.specs.instance_spec import LinkableInstanceSpec
from metricflow_semantics.specs.spec_set import group_specs_by_type
from metricflow_semantics.specs.spec_set_transforms import ToElementNameSet
from metricflow_semantics.specs.where_filter.where_filter_spec import WhereFilterSpec
from metricflow_semantics.sql.sql_join_type import SqlJoinType

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.builder.partitions import PartitionJoinResolver
from metricflow.dataflow.dataflow_plan import (
    DataflowPlanNode,
)
from metricflow.dataflow.nodes.constrain_time import ConstrainTimeRangeNode
from metricflow.dataflow.nodes.filter_elements import FilterElementsNode
from metricflow.dataflow.nodes.join_to_base import JoinDescription, JoinOnEntitiesNode
from metricflow.dataflow.nodes.metric_time_transform import MetricTimeDimensionTransformNode
from metricflow.dataflow.nodes.where_filter import WhereConstraintNode
from metricflow.validation.dataflow_join_validator import JoinDataflowOutputValidator

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class MultiHopJoinCandidateLineage:
    """Describes how the multi-hop join candidate was formed.

    For example, if
    * bridge_source has the primary entity account_id and the foreign entity customer_id
    * customers_source has the primary entity customer_id and dimension country
    * transactions_source has the transactions measure and the account_id foreign entity

    Then the candidate lineage can be describe as
        first_node_to_join = bridge_source
        second_node_to_join = customers_source
    to get the country dimension.
    """

    first_node_to_join: DataflowPlanNode
    second_node_to_join: DataflowPlanNode
    join_second_node_by_entity: LinklessEntitySpec


@dataclasses.dataclass(frozen=True)
class MultiHopJoinCandidate:
    """A candidate node containing linkable specs that is join of other nodes. It's used to resolve multi-hop queries.

    Also see MultiHopJoinCandidateLineage.
    """

    node_with_multi_hop_elements: DataflowPlanNode
    lineage: MultiHopJoinCandidateLineage


class PredicateInputType(Enum):
    """Enumeration of predicate input types we may encounter in where filters.

    This is primarily used for describing when predicate pushdown may operate in a dataflow plan, and is necessary
    for holistic checks against the set of potentially enabled pushdown operations. For example, in the scenario
    scenario where we only allow time range updates, we must do careful overriding of other pushdown properties.

    This also allows us to disable pushdown for things like time dimension filters in cases where we might
    accidentally censor input data.
    """

    CATEGORICAL_DIMENSION = "categorical_dimension"
    ENTITY = "entity"
    TIME_DIMENSION = "time_dimension"
    TIME_RANGE_CONSTRAINT = "time_range_constraint"


@dataclasses.dataclass(frozen=True)
class PredicatePushdownState:
    """Container class for maintaining state information relevant for predicate pushdown.

    This broadly tracks three related items:
    1. Filter predicates collected during the process of constructing a dataflow plan
    2. Predicate types eligible for pushdown
    3. Filters which have been applied already

    The first may be updated as things like time constraints get altered or metric and measure filters are
    added to the query filters.
    The second may be updated based on query configuration, like if a cumulative metric is added to the plan
    there may be changes to what sort of predicate pushdown operations are supported.
    The last will be updated as filters are applied via pushdown or by the original WhereConstraintNode.

    Finally, the time_range_constraint property holds the time window for setting up a time range filter expression.
    """

    time_range_constraint: Optional[TimeRangeConstraint]
    # TODO: Deduplicate where_filter_specs
    where_filter_specs: Tuple[WhereFilterSpec, ...]
    applied_where_filter_specs: FrozenSet[WhereFilterSpec] = frozenset()
    pushdown_enabled_types: FrozenSet[PredicateInputType] = frozenset(
        [PredicateInputType.TIME_RANGE_CONSTRAINT, PredicateInputType.CATEGORICAL_DIMENSION]
    )

    def __post_init__(self) -> None:
        """Validation to ensure pushdown states are configured correctly.

        In particular, this asserts that cases where pushdown is disabled cannot leak pushdown operations via
        outside property access - if pushdown is disabled, no further pushdown operations of any kind are allowed
        on that particular code branch. It also asserts that unsupported pushdown scenarios are not configured.
        """
        invalid_types: Set[PredicateInputType] = set()

        for input_type in self.pushdown_enabled_types:
            if input_type is PredicateInputType.ENTITY or input_type is PredicateInputType.TIME_DIMENSION:
                invalid_types.add(input_type)
            elif (
                input_type is PredicateInputType.CATEGORICAL_DIMENSION
                or input_type is PredicateInputType.TIME_RANGE_CONSTRAINT
            ):
                continue
            else:
                assert_values_exhausted(input_type)

        assert len(invalid_types) == 0, (
            "Unsupported predicate input type found in pushdown state configuration! We currently only support "
            "predicate pushdown for a subset of possible predicate input types (i.e., types of semantic manifest "
            "elements, such as entities and time dimensions, referenced in filter predicates), but this was enabled "
            f"for {self.pushdown_enabled_types}, which includes the following invalid types: {invalid_types}."
        )

        time_range_constraint_is_valid = (
            self.time_range_constraint is None
            or PredicateInputType.TIME_RANGE_CONSTRAINT in self.pushdown_enabled_types
        )
        where_filter_specs_are_valid = len(self.where_filter_specs) == 0 or self.where_filter_pushdown_enabled
        assert time_range_constraint_is_valid and where_filter_specs_are_valid, (
            "Invalid pushdown state configuration! Disabled pushdown state objects cannot have properties "
            "set that may lead to improper access and use in other contexts, as that can lead to unintended "
            "filtering operations in cases where these properties are accessed without appropriate checks against "
            "pushdown configuration. The following properties should be None or empty:\n"
            f"time_range_constraint: {self.time_range_constraint}\n"
            f"where_filter_specs: {self.where_filter_specs}"
        )

    @property
    def has_pushdown_potential(self) -> bool:
        """Returns whether or not pushdown is enabled for a type with predicate candidates in place."""
        return self.has_time_range_constraint_to_push_down or self.has_where_filters_to_push_down

    @property
    def has_time_range_constraint_to_push_down(self) -> bool:
        """Convenience accessor for checking if there is a time range constraint that can be pushed down.

        Note: this time range enabled state is a backwards compatibility shim for use with conversion metrics while
        we determine how best to support predicate pushdown for conversion metrics. It may have longer term utility,
        but ideally we'd collapse this with the more general time dimension filter input scenarios.
        """
        return (
            PredicateInputType.TIME_RANGE_CONSTRAINT in self.pushdown_enabled_types
            and self.time_range_constraint is not None
        )

    @property
    def has_where_filters_to_push_down(self) -> bool:
        """Convenience accessor for checking if there are any where filters to push down."""
        return self.where_filter_pushdown_enabled and len(self.where_filter_specs) > 0

    @property
    def where_filter_pushdown_enabled(self) -> bool:
        """Indicates whether or not pushdown is enabled for where filters."""
        return (
            PredicateInputType.CATEGORICAL_DIMENSION in self.pushdown_enabled_types
            or PredicateInputType.ENTITY in self.pushdown_enabled_types
            or PredicateInputType.TIME_DIMENSION in self.pushdown_enabled_types
        )

    @property
    def pushdown_eligible_element_types(self) -> FrozenSet[LinkableElementType]:
        """Set of linkable element types eligible for predicate pushdown.

        This converts from enabled PushdownInputTypes for checking if linkable elements in where filter specs are
        eligible for pushdown.
        """
        eligible_types: List[LinkableElementType] = []
        for enabled_type in self.pushdown_enabled_types:
            if enabled_type is PredicateInputType.TIME_RANGE_CONSTRAINT:
                pass
            elif enabled_type is PredicateInputType.CATEGORICAL_DIMENSION:
                eligible_types.append(LinkableElementType.DIMENSION)
            elif enabled_type is PredicateInputType.TIME_DIMENSION or enabled_type is PredicateInputType.ENTITY:
                # TODO: Remove as support for time dimensions and entities becomes available
                raise NotImplementedError(
                    "Predicate pushdown is not currently supported for where filter predicates with time dimension or "
                    f"entity references, but this pushdown state is enabled for {enabled_type}."
                )
            else:
                assert_values_exhausted(enabled_type)

        return frozenset(eligible_types)

    @staticmethod
    def with_time_range_constraint(
        original_pushdown_state: PredicatePushdownState, time_range_constraint: TimeRangeConstraint
    ) -> PredicatePushdownState:
        """Factory method for updating a pushdown state with a time range constraint.

        This allows for temporarily overriding a time range constraint with an adjusted one, or enabling a time
        range constraint filter if one becomes available mid-stream during dataflow plan construction.
        """
        pushdown_enabled_types = original_pushdown_state.pushdown_enabled_types.union(
            {PredicateInputType.TIME_RANGE_CONSTRAINT}
        )
        return PredicatePushdownState(
            time_range_constraint=time_range_constraint,
            pushdown_enabled_types=pushdown_enabled_types,
            where_filter_specs=original_pushdown_state.where_filter_specs,
            applied_where_filter_specs=original_pushdown_state.applied_where_filter_specs,
        )

    @staticmethod
    def without_time_range_constraint(
        original_pushdown_state: PredicatePushdownState,
    ) -> PredicatePushdownState:
        """Factory method for updating pushdown state to bypass time range constraints.

        This eliminates time range constraint pushdown as an option, since the only reason to remove
        time range constraint metadata is to turn it off, so we avoid potential issues where
        a second ConstrainTimeRange node might update the pushdown state.

        TODO: replace or rename this method.
        """
        pushdown_enabled_types = original_pushdown_state.pushdown_enabled_types.difference(
            {PredicateInputType.TIME_RANGE_CONSTRAINT}
        )
        return PredicatePushdownState(
            time_range_constraint=None,
            pushdown_enabled_types=pushdown_enabled_types,
            where_filter_specs=original_pushdown_state.where_filter_specs,
            applied_where_filter_specs=original_pushdown_state.applied_where_filter_specs,
        )

    @staticmethod
    def without_where_filter_specs(
        original_pushdown_state: PredicatePushdownState,
    ) -> PredicatePushdownState:
        """Factory method for updating pushdown state to remove existing where filter specs.

        This simply blanks out the where filter specs without altering which types of pushdown are available.
        """
        return PredicatePushdownState.with_where_filter_specs(
            original_pushdown_state=original_pushdown_state,
            where_filter_specs=tuple(),
        )

    @staticmethod
    def with_where_filter_specs(
        original_pushdown_state: PredicatePushdownState, where_filter_specs: Sequence[WhereFilterSpec]
    ) -> PredicatePushdownState:
        """Factory method for replacing WhereFilterSpecs in pushdown operations.

        This requires that the PushdownState allow for where filters - time range only or disabled states will raise
        an exception, and must be checked externally.
        """
        return PredicatePushdownState(
            time_range_constraint=original_pushdown_state.time_range_constraint,
            where_filter_specs=tuple(where_filter_specs),
            pushdown_enabled_types=original_pushdown_state.pushdown_enabled_types,
            applied_where_filter_specs=original_pushdown_state.applied_where_filter_specs,
        )

    @staticmethod
    def with_pushdown_applied_where_filter_specs(
        original_pushdown_state: PredicatePushdownState, pushdown_applied_where_filter_specs: FrozenSet[WhereFilterSpec]
    ) -> PredicatePushdownState:
        """Factory method for replacing pushdown applied where filter specs in pushdown operations.

        This is useful for managing propagation - both forwards and backwards - of where filter specs that have been
        applied via a pushdown operation.
        """
        return PredicatePushdownState(
            time_range_constraint=original_pushdown_state.time_range_constraint,
            pushdown_enabled_types=original_pushdown_state.pushdown_enabled_types,
            where_filter_specs=original_pushdown_state.where_filter_specs,
            applied_where_filter_specs=pushdown_applied_where_filter_specs,
        )

    @staticmethod
    def with_pushdown_disabled() -> PredicatePushdownState:
        """Factory method for configuring a disabled predicate pushdown state.

        This is useful in cases where there is a branched path where pushdown should be disabled in one branch while the
        other may remain eligible. For example, a join linkage where one side of the join contains an unsupported
        configuration might send a disabled copy of the pushdown parameters down that path while retaining the potential
        for using another path.
        """
        return PredicatePushdownState(
            time_range_constraint=None,
            pushdown_enabled_types=frozenset(),
            where_filter_specs=tuple(),
        )


class PreJoinNodeProcessor:
    """Processes source nodes before other nodes are joined.

    Generally, the source nodes will be combined with other dataflow plan nodes to produce a new set of nodes to realize
    a condition of the query. For example, to realize a time range constraint, a ConstrainTimeRangeNode will be added
    to the source nodes.

    e.g.

    <SomeDataflowPlanNode/>

    ->

    <ConstrainTimeRangeNode>
        <SomeDataflowPlanNode/>
    </ConstrainTimeRangeNode>

    """

    def __init__(  # noqa: D107
        self,
        semantic_model_lookup: SemanticModelLookup,
        node_data_set_resolver: DataflowPlanNodeOutputDataSetResolver,
    ):
        self._node_data_set_resolver = node_data_set_resolver
        self._partition_resolver = PartitionJoinResolver(semantic_model_lookup)
        self._semantic_model_lookup = semantic_model_lookup
        self._join_evaluator = JoinDataflowOutputValidator(semantic_model_lookup)

    def apply_matching_filter_predicates(
        self,
        source_nodes: Sequence[DataflowPlanNode],
        predicate_pushdown_state: PredicatePushdownState,
        metric_time_dimension_reference: TimeDimensionReference,
    ) -> Sequence[DataflowPlanNode]:
        """Adds filter predicate nodes to the input nodes as appropriate."""
        if predicate_pushdown_state.has_time_range_constraint_to_push_down:
            source_nodes = self._add_time_range_constraint(
                source_nodes=source_nodes,
                metric_time_dimension_reference=metric_time_dimension_reference,
                time_range_constraint=predicate_pushdown_state.time_range_constraint,
            )

        if predicate_pushdown_state.has_where_filters_to_push_down:
            source_nodes = self._add_where_constraint(
                source_nodes=source_nodes,
                where_filter_specs=predicate_pushdown_state.where_filter_specs,
                enabled_element_types=predicate_pushdown_state.pushdown_eligible_element_types,
            )

        return source_nodes

    def _add_time_range_constraint(
        self,
        source_nodes: Sequence[DataflowPlanNode],
        metric_time_dimension_reference: TimeDimensionReference,
        time_range_constraint: Optional[TimeRangeConstraint] = None,
    ) -> Sequence[DataflowPlanNode]:
        """Adds a time range constraint node to the input nodes."""
        processed_nodes: List[DataflowPlanNode] = []
        for source_node in source_nodes:
            # Constrain the time range if specified.
            if time_range_constraint:
                node_output_data_set = self._node_data_set_resolver.get_output_data_set(source_node)
                constrain_time = False
                for time_dimension_instance in node_output_data_set.instance_set.time_dimension_instances:
                    if (
                        time_dimension_instance.spec.reference == metric_time_dimension_reference
                        and len(time_dimension_instance.spec.entity_links) == 0
                    ):
                        constrain_time = True
                        break
                if constrain_time:
                    processed_nodes.append(
                        ConstrainTimeRangeNode.create(
                            parent_node=source_node, time_range_constraint=time_range_constraint
                        )
                    )
                else:
                    processed_nodes.append(source_node)
            else:
                processed_nodes.append(source_node)
        return processed_nodes

    def _add_where_constraint(
        self,
        source_nodes: Sequence[DataflowPlanNode],
        where_filter_specs: Sequence[WhereFilterSpec],
        enabled_element_types: FrozenSet[LinkableElementType],
    ) -> Sequence[DataflowPlanNode]:
        """Processes where filter specs and evaluates their fitness for pushdown against the provided node set."""
        eligible_filter_specs_by_model: Dict[SemanticModelReference, Sequence[WhereFilterSpec]] = {}
        for spec in where_filter_specs:
            semantic_models = set(element.semantic_model_origin for element in spec.linkable_elements)
            invalid_element_types = [
                element for element in spec.linkable_elements if element.element_type not in enabled_element_types
            ]
            if len(semantic_models) == 1 and len(invalid_element_types) == 0:
                model = semantic_models.pop()
                eligible_filter_specs_by_model[model] = tuple(eligible_filter_specs_by_model.get(model, tuple())) + (
                    spec,
                )

        filtered_nodes: List[DataflowPlanNode] = []
        for source_node in source_nodes:
            node_semantic_models = tuple(source_node.as_plan().source_semantic_models)
            if len(node_semantic_models) == 1 and node_semantic_models[0] in eligible_filter_specs_by_model:
                eligible_filter_specs = eligible_filter_specs_by_model[node_semantic_models[0]]
                source_node_specs = self._node_data_set_resolver.get_output_data_set(source_node).instance_set.spec_set
                matching_filter_specs = [
                    filter_spec
                    for filter_spec in eligible_filter_specs
                    if all([spec in source_node_specs.linkable_specs for spec in filter_spec.linkable_specs])
                ]
                if len(matching_filter_specs) == 0:
                    filtered_nodes.append(source_node)
                else:
                    filtered_nodes.append(
                        WhereConstraintNode.create(parent_node=source_node, where_specs=matching_filter_specs)
                    )
            else:
                filtered_nodes.append(source_node)

        return filtered_nodes

    def _node_contains_entity(
        self,
        node: DataflowPlanNode,
        entity_reference: EntityReference,
    ) -> bool:
        """Returns true if the output of the node contains an entity of the given types."""
        data_set = self._node_data_set_resolver.get_output_data_set(node)

        for entity_instance_in_first_node in data_set.instance_set.entity_instances:
            entity_spec_in_first_node = entity_instance_in_first_node.spec

            if entity_spec_in_first_node.reference != entity_reference:
                continue

            assert (
                len(entity_instance_in_first_node.defined_from) == 1
            ), "Multiple items in defined_from not yet supported"

            entity = self._semantic_model_lookup.get_entity_in_semantic_model(
                entity_instance_in_first_node.defined_from[0]
            )
            if entity is None:
                raise RuntimeError(
                    f"Invalid SemanticModelElementReference {entity_instance_in_first_node.defined_from[0]}"
                )

            return True

        return False

    def _get_candidates_nodes_for_multi_hop(
        self, desired_linkable_spec: LinkableInstanceSpec, nodes: Sequence[DataflowPlanNode], join_type: SqlJoinType
    ) -> Sequence[MultiHopJoinCandidate]:
        """Assemble nodes representing all possible one-hop joins."""
        if len(desired_linkable_spec.entity_links) > MAX_JOIN_HOPS:
            raise NotImplementedError(
                f"Multi-hop joins with more than {MAX_JOIN_HOPS} entity links not yet supported. "
                f"Got: {desired_linkable_spec}"
            )
        if len(desired_linkable_spec.entity_links) != 2:
            return ()

        multi_hop_join_candidates: List[MultiHopJoinCandidate] = []
        logger.debug(LazyFormat(lambda: f"Creating nodes for {desired_linkable_spec}"))

        for first_node_that_could_be_joined in nodes:
            data_set_of_first_node_that_could_be_joined = self._node_data_set_resolver.get_output_data_set(
                first_node_that_could_be_joined
            )

            # When joining on the entity, the first node needs the first and second entity links.
            if not (
                self._node_contains_entity(
                    node=first_node_that_could_be_joined,
                    entity_reference=desired_linkable_spec.entity_links[0],
                )
                and self._node_contains_entity(
                    node=first_node_that_could_be_joined,
                    entity_reference=desired_linkable_spec.entity_links[1],
                )
            ):
                continue

            for second_node_that_could_be_joined in nodes:
                if not (
                    self._node_contains_entity(
                        node=second_node_that_could_be_joined,
                        entity_reference=desired_linkable_spec.entity_links[1],
                    )
                ):
                    continue

                # Avoid loops between the same semantic models.
                if second_node_that_could_be_joined.node_id == first_node_that_could_be_joined.node_id:
                    continue

                data_set_of_second_node_that_can_be_joined = self._node_data_set_resolver.get_output_data_set(
                    second_node_that_could_be_joined
                )

                # If the element name of the linkable spec doesn't exist in the joined data set, then it can't be
                # useful for obtaining that linkable spec.
                element_names_in_data_set = ToElementNameSet().transform(
                    data_set_of_second_node_that_can_be_joined.instance_set.spec_set
                )

                if desired_linkable_spec.element_name not in element_names_in_data_set:
                    continue

                # The first and second nodes are joined by this entity
                entity_reference_to_join_first_and_second_nodes = desired_linkable_spec.entity_links[1]

                if not self._join_evaluator.is_valid_instance_set_join(
                    left_instance_set=data_set_of_first_node_that_could_be_joined.instance_set,
                    right_instance_set=data_set_of_second_node_that_can_be_joined.instance_set,
                    on_entity_reference=entity_reference_to_join_first_and_second_nodes,
                    right_node_is_aggregated_to_entity=second_node_that_could_be_joined.aggregated_to_elements
                    == {entity_reference_to_join_first_and_second_nodes},
                ):
                    continue

                # filter measures out of joinable_node
                specs = data_set_of_second_node_that_can_be_joined.instance_set.spec_set
                filtered_joinable_node = FilterElementsNode.create(
                    parent_node=second_node_that_could_be_joined,
                    include_specs=group_specs_by_type(
                        specs.dimension_specs
                        + specs.entity_specs
                        + specs.time_dimension_specs
                        + specs.group_by_metric_specs
                    ),
                )

                join_on_partition_dimensions = self._partition_resolver.resolve_partition_dimension_joins(
                    left_node_spec_set=data_set_of_first_node_that_could_be_joined.instance_set.spec_set,
                    node_to_join_spec_set=data_set_of_second_node_that_can_be_joined.instance_set.spec_set,
                )
                join_on_partition_time_dimensions = self._partition_resolver.resolve_partition_time_dimension_joins(
                    left_node_spec_set=data_set_of_first_node_that_could_be_joined.instance_set.spec_set,
                    node_to_join_spec_set=data_set_of_second_node_that_can_be_joined.instance_set.spec_set,
                )

                multi_hop_join_candidates.append(
                    MultiHopJoinCandidate(
                        node_with_multi_hop_elements=JoinOnEntitiesNode.create(
                            left_node=first_node_that_could_be_joined,
                            join_targets=[
                                JoinDescription(
                                    join_node=filtered_joinable_node,
                                    join_on_entity=LinklessEntitySpec.from_reference(
                                        desired_linkable_spec.entity_links[1]
                                    ),
                                    join_on_partition_dimensions=join_on_partition_dimensions,
                                    join_on_partition_time_dimensions=join_on_partition_time_dimensions,
                                    join_type=join_type,
                                )
                            ],
                        ),
                        lineage=MultiHopJoinCandidateLineage(
                            first_node_to_join=first_node_that_could_be_joined,
                            second_node_to_join=second_node_that_could_be_joined,
                            # entity_spec_in_first_node should already not have entity links since we checked
                            # for that, but using this method for type checking.
                            join_second_node_by_entity=LinklessEntitySpec.from_reference(
                                desired_linkable_spec.entity_links[1]
                            ),
                        ),
                    )
                )

        for multi_hop_join_candidate in multi_hop_join_candidates:
            output_data_set = self._node_data_set_resolver.get_output_data_set(
                multi_hop_join_candidate.node_with_multi_hop_elements
            )
            logger.debug(
                LazyFormat(
                    lambda: f"Node {multi_hop_join_candidate.node_with_multi_hop_elements} has spec set:\n"
                    f"{mf_pformat(output_data_set.instance_set.spec_set)}"
                )
            )

        return multi_hop_join_candidates

    def add_multi_hop_joins(
        self,
        desired_linkable_specs: Sequence[LinkableInstanceSpec],
        nodes: Sequence[DataflowPlanNode],
        join_type: SqlJoinType,
    ) -> Sequence[DataflowPlanNode]:
        """Assemble nodes representing all possible one-hop joins."""
        all_multi_hop_join_candidates: List[MultiHopJoinCandidate] = []
        lineage_for_all_multi_hop_join_candidates: Set[MultiHopJoinCandidateLineage] = set()

        for desired_linkable_spec in desired_linkable_specs:
            for multi_hop_join_candidate in self._get_candidates_nodes_for_multi_hop(
                desired_linkable_spec=desired_linkable_spec, nodes=nodes, join_type=join_type
            ):
                # Dedupe candidates that are the same join.
                if multi_hop_join_candidate.lineage not in lineage_for_all_multi_hop_join_candidates:
                    all_multi_hop_join_candidates.append(multi_hop_join_candidate)
                    lineage_for_all_multi_hop_join_candidates.add(multi_hop_join_candidate.lineage)

        return list(x.node_with_multi_hop_elements for x in all_multi_hop_join_candidates) + list(nodes)

    def remove_unnecessary_nodes(
        self,
        desired_linkable_specs: Sequence[LinkableInstanceSpec],
        nodes: Sequence[DataflowPlanNode],
        metric_time_dimension_reference: TimeDimensionReference,
        time_spine_metric_time_nodes: Sequence[MetricTimeDimensionTransformNode],
    ) -> List[DataflowPlanNode]:
        """Filters out many of the nodes that can't possibly be useful for joins to obtain the desired linkable specs.

        A simple filter is to remove any nodes that don't share a common element with the query. Having a common element
        doesn't mean that the node will be useful, but not having common elements definitely means it's not useful.
        """
        relevant_element_names = {x.element_name for x in desired_linkable_specs}.union(
            {y.element_name for x in desired_linkable_specs for y in x.entity_links}
        )

        # The metric time dimension is used everywhere, so don't count it unless specifically desired in linkable spec
        # that has entity links.
        metric_time_dimension_used_in_linked_spec = any(
            [
                len(linkable_spec.entity_links) > 0
                and linkable_spec.element_name == metric_time_dimension_reference.element_name
                for linkable_spec in desired_linkable_specs
            ]
        )

        if (
            metric_time_dimension_reference.element_name in relevant_element_names
            and not metric_time_dimension_used_in_linked_spec
        ):
            relevant_element_names.remove(metric_time_dimension_reference.element_name)

        logger.debug(LazyFormat(lambda: f"Relevant names are: {relevant_element_names}"))

        relevant_nodes = []

        for node in nodes:
            logger.debug(LazyFormat(lambda: f"Examining {node} for pruning"))
            data_set = self._node_data_set_resolver.get_output_data_set(node)
            element_names_in_data_set = ToElementNameSet().transform(data_set.instance_set.spec_set)
            element_names_intersection = element_names_in_data_set.intersection(relevant_element_names)
            if len(element_names_intersection) > 0:
                logger.debug(
                    LazyFormat(
                        lambda: f"Including {node} since `element_names_intersection` is {element_names_intersection}"
                    )
                )
                relevant_nodes.append(node)
                continue

            # Used for group-by-item-values queries.
            if node in time_spine_metric_time_nodes:
                logger.debug(LazyFormat(lambda: f"Including {node} since it matches `time_spine_node`"))
                relevant_nodes.append(node)
                continue

        return relevant_nodes



from __future__ import annotations

import logging
import re
from typing import Optional

from dbt_semantic_interfaces.call_parameter_sets import (
    ParseWhereFilterException,
)
from dbt_semantic_interfaces.implementations.filters.where_filter import PydanticWhereFilter
from dbt_semantic_interfaces.parsing.where_filter.where_filter_parser import WhereFilterParser
from dbt_semantic_interfaces.references import EntityReference
from typing_extensions import override

from metricflow_semantics.model.semantic_manifest_lookup import SemanticManifestLookup
from metricflow_semantics.naming.naming_scheme import QueryItemNamingScheme
from metricflow_semantics.naming.object_builder_str import ObjectBuilderNameConverter
from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.patterns.entity_link_pattern import (
    EntityLinkPattern,
    ParameterSetField,
    SpecPatternParameterSet,
)
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.patterns.typed_patterns import DimensionPattern, TimeDimensionPattern

logger = logging.getLogger(__name__)


class ObjectBuilderNamingScheme(QueryItemNamingScheme):
    """A naming scheme using a builder syntax like Dimension('metric_time').grain('day')."""

    _NAME_REGEX = re.compile(r"\A(Dimension|TimeDimension|Entity|Metric)\(.*\)\Z")

    @override
    def input_str(self, instance_spec: InstanceSpec) -> Optional[str]:
        return ObjectBuilderNameConverter.input_str_from_spec(instance_spec)

    @override
    def spec_pattern(self, input_str: str, semantic_manifest_lookup: SemanticManifestLookup) -> SpecPattern:
        if not self.input_str_follows_scheme(input_str):
            raise ValueError(
                f"The specified input {repr(input_str)} does not match the input described by the object builder "
                f"pattern."
            )
        try:
            # TODO: Update when more appropriate parsing libraries are available.
            call_parameter_sets = PydanticWhereFilter(where_sql_template="{{ " + input_str + " }}").call_parameter_sets
        except ParseWhereFilterException as e:
            raise ValueError(f"A spec pattern can't be generated from the input string {repr(input_str)}") from e

        num_parameter_sets = (
            len(call_parameter_sets.dimension_call_parameter_sets)
            + len(call_parameter_sets.time_dimension_call_parameter_sets)
            + len(call_parameter_sets.entity_call_parameter_sets)
            + len(call_parameter_sets.metric_call_parameter_sets)
        )
        if num_parameter_sets != 1:
            raise ValueError(f"Did not find exactly 1 call parameter set. Got: {num_parameter_sets}")

        for dimension_call_parameter_set in call_parameter_sets.dimension_call_parameter_sets:
            return DimensionPattern(
                SpecPatternParameterSet.from_parameters(
                    element_name=dimension_call_parameter_set.dimension_reference.element_name,
                    entity_links=dimension_call_parameter_set.entity_path,
                    fields_to_compare=(
                        ParameterSetField.ELEMENT_NAME,
                        ParameterSetField.ENTITY_LINKS,
                        ParameterSetField.DATE_PART,
                    ),
                )
            )

        for time_dimension_call_parameter_set in call_parameter_sets.time_dimension_call_parameter_sets:
            fields_to_compare = [
                ParameterSetField.ELEMENT_NAME,
                ParameterSetField.ENTITY_LINKS,
                ParameterSetField.DATE_PART,
            ]

            if time_dimension_call_parameter_set.time_granularity_name is not None:
                fields_to_compare.append(ParameterSetField.TIME_GRANULARITY)

            return TimeDimensionPattern(
                SpecPatternParameterSet.from_parameters(
                    element_name=time_dimension_call_parameter_set.time_dimension_reference.element_name,
                    entity_links=time_dimension_call_parameter_set.entity_path,
                    time_granularity_name=time_dimension_call_parameter_set.time_granularity_name,
                    date_part=time_dimension_call_parameter_set.date_part,
                    fields_to_compare=tuple(fields_to_compare),
                )
            )

        for entity_call_parameter_set in call_parameter_sets.entity_call_parameter_sets:
            return EntityLinkPattern(
                SpecPatternParameterSet.from_parameters(
                    element_name=entity_call_parameter_set.entity_reference.element_name,
                    entity_links=entity_call_parameter_set.entity_path,
                    fields_to_compare=(
                        ParameterSetField.ELEMENT_NAME,
                        ParameterSetField.ENTITY_LINKS,
                    ),
                )
            )

        for metric_call_parameter_set in call_parameter_sets.metric_call_parameter_sets:
            return EntityLinkPattern(
                SpecPatternParameterSet.from_parameters(
                    element_name=metric_call_parameter_set.metric_reference.element_name,
                    entity_links=tuple(
                        EntityReference(element_name=group_by_ref.element_name)
                        for group_by_ref in metric_call_parameter_set.group_by
                    ),
                    fields_to_compare=(
                        ParameterSetField.ELEMENT_NAME,
                        ParameterSetField.ENTITY_LINKS,
                    ),
                )
            )

        raise RuntimeError("There should have been a return associated with one of the CallParameterSets.")

    @override
    def input_str_follows_scheme(self, input_str: str) -> bool:
        if ObjectBuilderNamingScheme._NAME_REGEX.match(input_str) is None:
            return False
        try:
            call_parameter_sets = WhereFilterParser.parse_call_parameter_sets("{{ " + input_str + " }}")
            return_value = (
                len(call_parameter_sets.dimension_call_parameter_sets)
                + len(call_parameter_sets.time_dimension_call_parameter_sets)
                + len(call_parameter_sets.entity_call_parameter_sets)
                + len(call_parameter_sets.metric_call_parameter_sets)
            ) == 1
            return return_value
        except ParseWhereFilterException:
            return False

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id()={hex(id(self))})"



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Mapping, Optional, Sequence, Tuple, Union

from dbt_semantic_interfaces.naming.keywords import METRIC_TIME_ELEMENT_NAME
from dbt_semantic_interfaces.references import DimensionReference, EntityReference, TimeDimensionReference
from dbt_semantic_interfaces.type_enums import DatePart, TimeGranularity
from typing_extensions import override

from metricflow_semantics.aggregation_properties import AggregationState
from metricflow_semantics.collection_helpers.lru_cache import typed_lru_cache
from metricflow_semantics.model.semantics.linkable_element import ElementPathKey, LinkableElementType
from metricflow_semantics.naming.linkable_spec_name import StructuredLinkableSpecName
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.instance_spec import InstanceSpecVisitor
from metricflow_semantics.sql.sql_exprs import SqlWindowFunction
from metricflow_semantics.time.granularity import ExpandedTimeGranularity
from metricflow_semantics.visitor import VisitorOutputT


class TimeDimensionSpecField(Enum):
    """Fields of the time dimension spec.

    The value corresponds to the name of the field in the dataclass. This should contain all fields, but implementation
    is pending.
    """

    TIME_GRANULARITY = "time_granularity"


class TimeDimensionSpecComparisonKey:
    """A key that can be used for comparing / grouping time dimension specs.

    Useful for assessing if two time dimension specs are equal while ignoring specific attributes. Keys must have the
    same set of excluded attributes to be valid for comparison.

    This is useful for ambiguous group-by-item resolution where we want to select a time dimension regardless of the
    grain.
    """

    def __init__(self, source_spec: TimeDimensionSpec, exclude_fields: Sequence[TimeDimensionSpecField]) -> None:
        """Initializer.

        Args:
            source_spec: The spec that this key is based on.
            exclude_fields: The fields to ignore when determining equality.
        """
        self._excluded_fields = frozenset(exclude_fields)
        self._source_spec = source_spec

        # This is a list of field values of TimeDimensionSpec that we should use for comparison.
        spec_field_values_for_comparison: List[
            Union[str, Tuple[EntityReference, ...], ExpandedTimeGranularity, Optional[DatePart]]
        ] = [self._source_spec.element_name, self._source_spec.entity_links]

        if TimeDimensionSpecField.TIME_GRANULARITY not in self._excluded_fields:
            spec_field_values_for_comparison.append(self._source_spec.time_granularity)

        spec_field_values_for_comparison.append(self._source_spec.date_part)

        self._spec_field_values_for_comparison = tuple(spec_field_values_for_comparison)

    @property
    def source_spec(self) -> TimeDimensionSpec:  # noqa: D102
        return self._source_spec

    @override
    def __eq__(self, other: Any) -> bool:  # type: ignore[misc]
        if not isinstance(other, TimeDimensionSpecComparisonKey):
            return False

        if self._excluded_fields != other._excluded_fields:
            return False

        return self._spec_field_values_for_comparison == other._spec_field_values_for_comparison

    @override
    def __hash__(self) -> int:
        return hash((self._excluded_fields, self._spec_field_values_for_comparison))


DEFAULT_TIME_GRANULARITY = TimeGranularity.DAY


@dataclass(frozen=True)
class TimeDimensionSpec(DimensionSpec):  # noqa: D101
    time_granularity: ExpandedTimeGranularity = ExpandedTimeGranularity.from_time_granularity(DEFAULT_TIME_GRANULARITY)
    date_part: Optional[DatePart] = None

    # Used for semi-additive joins. Some more thought is needed, but this may be useful in InstanceSpec.
    aggregation_state: Optional[AggregationState] = None

    window_function: Optional[SqlWindowFunction] = None

    @property
    def without_first_entity_link(self) -> TimeDimensionSpec:  # noqa: D102
        assert len(self.entity_links) > 0, f"Spec does not have any entity links: {self}"
        return TimeDimensionSpec(
            element_name=self.element_name,
            entity_links=self.entity_links[1:],
            time_granularity=self.time_granularity,
            date_part=self.date_part,
            aggregation_state=self.aggregation_state,
            window_function=self.window_function,
        )

    @property
    def without_entity_links(self) -> TimeDimensionSpec:  # noqa: D102
        return TimeDimensionSpec(
            element_name=self.element_name,
            time_granularity=self.time_granularity,
            date_part=self.date_part,
            entity_links=(),
            aggregation_state=self.aggregation_state,
            window_function=self.window_function,
        )

    @property
    def reference(self) -> TimeDimensionReference:  # noqa: D102
        return TimeDimensionReference(element_name=self.element_name)

    @property
    def dimension_reference(self) -> DimensionReference:  # noqa: D102
        return DimensionReference(element_name=self.element_name)

    @property
    def qualified_name(self) -> str:  # noqa: D102
        return StructuredLinkableSpecName(
            entity_link_names=tuple(x.element_name for x in self.entity_links),
            element_name=self.element_name,
            time_granularity_name=self.time_granularity.name,
            date_part=self.date_part,
        ).qualified_name

    @property
    @override
    def element_path_key(self) -> ElementPathKey:
        return ElementPathKey(
            element_name=self.element_name,
            element_type=LinkableElementType.TIME_DIMENSION,
            entity_links=self.entity_links,
            time_granularity=self.time_granularity,
            date_part=self.date_part,
        )

    @staticmethod
    def from_reference(reference: TimeDimensionReference) -> TimeDimensionSpec:
        """Initialize from a time dimension reference instance."""
        return TimeDimensionSpec(entity_links=(), element_name=reference.element_name)

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_time_dimension_spec(self)

    def with_grain(self, time_granularity: ExpandedTimeGranularity) -> TimeDimensionSpec:  # noqa: D102
        return TimeDimensionSpec(
            element_name=self.element_name,
            entity_links=self.entity_links,
            time_granularity=time_granularity,
            date_part=self.date_part,
            aggregation_state=self.aggregation_state,
            window_function=self.window_function,
        )

    def with_base_grain(self) -> TimeDimensionSpec:  # noqa: D102
        return TimeDimensionSpec(
            element_name=self.element_name,
            entity_links=self.entity_links,
            time_granularity=ExpandedTimeGranularity.from_time_granularity(self.time_granularity.base_granularity),
            date_part=self.date_part,
            aggregation_state=self.aggregation_state,
            window_function=self.window_function,
        )

    def with_grain_and_date_part(  # noqa: D102
        self, time_granularity: ExpandedTimeGranularity, date_part: Optional[DatePart]
    ) -> TimeDimensionSpec:
        return TimeDimensionSpec(
            element_name=self.element_name,
            entity_links=self.entity_links,
            time_granularity=time_granularity,
            date_part=date_part,
            aggregation_state=self.aggregation_state,
            window_function=self.window_function,
        )

    def with_aggregation_state(self, aggregation_state: AggregationState) -> TimeDimensionSpec:  # noqa: D102
        return TimeDimensionSpec(
            element_name=self.element_name,
            entity_links=self.entity_links,
            time_granularity=self.time_granularity,
            date_part=self.date_part,
            aggregation_state=aggregation_state,
            window_function=self.window_function,
        )

    def with_window_function(self, window_function: SqlWindowFunction) -> TimeDimensionSpec:  # noqa: D102
        return TimeDimensionSpec(
            element_name=self.element_name,
            entity_links=self.entity_links,
            time_granularity=self.time_granularity,
            date_part=self.date_part,
            aggregation_state=self.aggregation_state,
            window_function=window_function,
        )

    def comparison_key(self, exclude_fields: Sequence[TimeDimensionSpecField] = ()) -> TimeDimensionSpecComparisonKey:
        """See TimeDimensionComparisonKey."""
        return TimeDimensionSpecComparisonKey(
            source_spec=self,
            exclude_fields=exclude_fields,
        )

    @classmethod
    @typed_lru_cache
    def _get_compatible_grain_and_date_part(cls) -> Sequence[Tuple[ExpandedTimeGranularity, DatePart]]:
        items = []
        for date_part in DatePart:
            for compatible_granularity in date_part.compatible_granularities:
                items.append((ExpandedTimeGranularity.from_time_granularity(compatible_granularity), date_part))
        return items

    @classmethod
    def generate_possible_specs_for_time_dimension(
        cls,
        time_dimension_reference: TimeDimensionReference,
        entity_links: Tuple[EntityReference, ...],
        custom_granularities: Mapping[str, ExpandedTimeGranularity],
    ) -> List[TimeDimensionSpec]:
        """Generate a list of time dimension specs with all combinations of granularity & date part."""
        time_dimension_specs: List[TimeDimensionSpec] = []
        granularities = [
            ExpandedTimeGranularity.from_time_granularity(time_granularity) for time_granularity in TimeGranularity
        ] + list(custom_granularities.values())
        for time_granularity in granularities:
            time_dimension_specs.append(
                TimeDimensionSpec(
                    element_name=time_dimension_reference.element_name,
                    entity_links=entity_links,
                    time_granularity=time_granularity,
                    date_part=None,
                )
            )
        for grain, date_part in cls._get_compatible_grain_and_date_part():
            time_dimension_specs.append(
                TimeDimensionSpec(
                    element_name=time_dimension_reference.element_name,
                    entity_links=entity_links,
                    time_granularity=grain,
                    date_part=date_part,
                )
            )
        return time_dimension_specs

    @property
    def is_metric_time(self) -> bool:  # noqa: D102
        return self.element_name == METRIC_TIME_ELEMENT_NAME

    def with_entity_prefix(self, entity_prefix: EntityReference) -> TimeDimensionSpec:  # noqa: D102
        return TimeDimensionSpec(
            element_name=self.element_name,
            entity_links=(entity_prefix,) + self.entity_links,
            time_granularity=self.time_granularity,
            date_part=self.date_part,
            aggregation_state=self.aggregation_state,
            window_function=self.window_function,
        )

    @staticmethod
    def with_base_grains(time_dimension_specs: Sequence[TimeDimensionSpec]) -> Tuple[TimeDimensionSpec, ...]:
        """Return the list of time dimension specs, replacing any custom grains with base grains.

        Dedupes new specs, but preserves the initial order.
        """
        base_grain_specs: Tuple[TimeDimensionSpec, ...] = ()
        for spec in time_dimension_specs:
            base_grain_spec = spec.with_base_grain()
            if base_grain_spec not in base_grain_specs:
                base_grain_specs += (base_grain_spec,)
        return base_grain_specs



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan import DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor


@dataclass(frozen=True, eq=False)
class AddGeneratedUuidColumnNode(DataflowPlanNode):
    """Adds a UUID column."""

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(parent_node: DataflowPlanNode) -> AddGeneratedUuidColumnNode:  # noqa: D102
        return AddGeneratedUuidColumnNode(parent_nodes=(parent_node,))

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_ADD_UUID_COLUMN_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_add_generated_uuid_column_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return "Adds an internally generated UUID column"

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return super().displayed_properties

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__)

    def with_new_parents(  # noqa: D102
        self, new_parent_nodes: Sequence[DataflowPlanNode]
    ) -> AddGeneratedUuidColumnNode:
        assert len(new_parent_nodes) == 1
        return AddGeneratedUuidColumnNode(parent_nodes=(new_parent_nodes[0],))



from __future__ import annotations

import difflib
import logging
import os
import pathlib
import re
import webbrowser
from dataclasses import dataclass
from typing import Any, Callable, Mapping, Optional, Tuple, TypeVar

import _pytest.fixtures
import tabulate
from _pytest.fixtures import FixtureRequest

from metricflow_semantics.dag.mf_dag import MetricFlowDag
from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.model.semantics.linkable_element_set import LinkableElementSet
from metricflow_semantics.naming.object_builder_scheme import ObjectBuilderNamingScheme
from metricflow_semantics.specs.linkable_spec_set import LinkableSpecSet
from metricflow_semantics.specs.spec_set import InstanceSpecSet

logger = logging.getLogger(__name__)

# In the snapshot file, include a header that describe what should be observed in the snapshot.
SNAPSHOT_EXPECTATION_DESCRIPTION = "expectation_description"


@dataclass(frozen=True)
class SnapshotConfiguration:
    """Configuration for handling snapshots in a test session."""

    # Whether to display the snapshot associated with a test session in a browser window.
    display_snapshots: bool
    # Whether to overwrite any text files that were generated.
    overwrite_snapshots: bool
    # Absolute directory where the snapshots should be stored.
    snapshot_directory: pathlib.Path
    # Absolute directory where the tests are stored.
    tests_directory: pathlib.Path


def assert_snapshot_text_equal(
    request: _pytest.fixtures.FixtureRequest,
    mf_test_configuration: SnapshotConfiguration,
    group_id: str,
    snapshot_id: str,
    snapshot_text: str,
    snapshot_file_extension: str,
    exclude_line_regex: Optional[str] = None,
    incomparable_strings_replacement_function: Optional[Callable[[str], str]] = None,
    additional_sub_directories_for_snapshots: Tuple[str, ...] = (),
    additional_header_fields: Optional[Mapping[str, str]] = None,
    expectation_description: Optional[str] = None,
) -> None:
    """Similar to assert_plan_snapshot_text_equal(), but with more controls on how the snapshot paths are generated."""
    logger.debug(LazyFormat(lambda: "Generated snapshot text:\n" + indent(snapshot_text)))
    file_path = (
        str(
            snapshot_path_prefix(
                request=request,
                snapshot_configuration=mf_test_configuration,
                snapshot_group=group_id,
                snapshot_id=snapshot_id,
                additional_sub_directories=additional_sub_directories_for_snapshots,
            )
        )
        + snapshot_file_extension
    )

    if incomparable_strings_replacement_function is not None:
        snapshot_text = incomparable_strings_replacement_function(snapshot_text)

    # Add a header with context about the snapshot.
    path_to_test_file = pathlib.Path(request.node.fspath)
    test_doc_string = request.function.__doc__
    header_lines = [
        f"test_name: {request.node.name}",
        f"test_filename: {path_to_test_file.name}",
    ]
    if test_doc_string is not None:
        header_lines.append("docstring:")
        header_lines.append(indent(test_doc_string.rstrip()))
    if additional_header_fields is not None:
        for header_field_name, header_field_value in additional_header_fields.items():
            header_lines.append(f"{header_field_name}: {header_field_value}")
    if expectation_description is not None:
        header_lines.append(f"{SNAPSHOT_EXPECTATION_DESCRIPTION}:")
        header_lines.append(indent(expectation_description))
    header_lines.append("---")

    snapshot_text = "\n".join(header_lines) + "\n" + snapshot_text

    # Add a new line at the end of the file so that PRs don't show the "no newline" symbol on Github.
    if len(snapshot_text) > 1 and snapshot_text[-1] != "\n":
        snapshot_text = snapshot_text + "\n"

    # If we are in overwrite mode, create / overwrite the snapshot file.:
    if mf_test_configuration.overwrite_snapshots:
        # Create parent directory for the plan text files.
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w") as snapshot_text_file:
            snapshot_text_file.write(snapshot_text)

    # Throw an exception if the plan is not there.
    if not os.path.exists(file_path):
        raise FileNotFoundError(
            f"Could not find snapshot file at path {file_path}. Re-run with --overwrite-snapshots and check git status "
            f"to see what's new."
        )

    if mf_test_configuration.display_snapshots:
        if not mf_test_configuration.overwrite_snapshots:
            logger.warning(
                LazyFormat(lambda: f"Not overwriting snapshots, so displaying existing snapshot at {file_path}")
            )

        if len(request.session.items) > 1:
            raise ValueError("Displaying snapshots is only supported when there's a single item in a testing session.")
        webbrowser.open("file://" + file_path)

    # Read the existing plan from the file and compare with the actual plan
    with open(file_path, "r") as snapshot_text_file:
        expected_snapshot_text = snapshot_text_file.read()

        if exclude_line_regex:
            # Filter out lines that should be ignored.
            expected_snapshot_text = _exclude_lines_matching_regex(
                file_contents=expected_snapshot_text, exclude_line_regex=exclude_line_regex
            )
            snapshot_text = _exclude_lines_matching_regex(
                file_contents=snapshot_text, exclude_line_regex=exclude_line_regex
            )
        # pytest should show a detailed diff with "assert actual_modified == expected_modified", but it's not, so doing
        # this instead.
        if snapshot_text != expected_snapshot_text:
            diff = difflib.unified_diff(
                a=expected_snapshot_text.splitlines(keepends=True),
                b=snapshot_text.splitlines(keepends=True),
                fromfile=f"Expected Result in {file_path}",
                tofile="Actual Result",
            )
            assert False, "Result does not match the stored snapshot. Diff from expected to actual:\n\n" + "".join(diff)


def snapshot_path_prefix(
    request: _pytest.fixtures.FixtureRequest,
    snapshot_configuration: SnapshotConfiguration,
    snapshot_group: str,
    snapshot_id: str,
    additional_sub_directories: Tuple[str, ...] = (),
) -> pathlib.Path:
    """Returns a path prefix that can be used to build filenames for files associated with the snapshot.

    The snapshot prefix is generated from the name of the test file, the name of the test, name of the snapshot class,
    and the name of the snapshot.

    e.g.
    .../snapshots/test_file.py/DataflowPlan/test_name__plan1

    which can be used to construct paths like

    .../snapshots/test_file.py/DataflowPlan/test_name__plan1.xml
    .../snapshots/test_file.py/DataflowPlan/test_name__plan1.svg
    """
    test_name = request.node.name

    snapshot_file_name_parts = []
    # Parameterized test names look like 'test_case[some_param]'. "[" and "]" are annoying to deal with in the shell,
    # so replace them with dunders.
    snapshot_file_name_parts.extend(re.split(r"[\[\]]", test_name))
    # A trailing ] will produce an empty string in the list, so remove that.
    snapshot_file_name_parts = [part for part in snapshot_file_name_parts if len(part) > 0]
    snapshot_file_name_parts.append(snapshot_id)

    snapshot_file_name_prefix = "__".join(snapshot_file_name_parts)
    path_to_test_file = pathlib.Path(request.node.fspath)
    directory_to_store_snapshot = snapshot_configuration.snapshot_directory.joinpath(
        snapshot_configuration.snapshot_directory,
        path_to_test_file.name,
        snapshot_group,
    )
    if additional_sub_directories is not None:
        directory_to_store_snapshot = directory_to_store_snapshot.joinpath(*additional_sub_directories)

    return directory_to_store_snapshot.joinpath(snapshot_file_name_prefix)


def _exclude_lines_matching_regex(file_contents: str, exclude_line_regex: str) -> str:
    """Removes lines from file contents if the line matches exclude_regex."""
    compiled_regex = re.compile(exclude_line_regex)
    return "\n".join([line for line in file_contents.split("\n") if not compiled_regex.match(line)])


DISPLAY_SNAPSHOTS_CLI_FLAG = "--display-snapshots"
OVERWRITE_SNAPSHOTS_CLI_FLAG = "--overwrite-snapshots"


def add_display_snapshots_cli_flag(parser: _pytest.config.argparsing.Parser) -> None:  # noqa: D103
    parser.addoption(DISPLAY_SNAPSHOTS_CLI_FLAG, action="store_true", help="Displays snapshots in a browser if set")


def add_overwrite_snapshots_cli_flag(parser: _pytest.config.argparsing.Parser) -> None:  # noqa: D103
    parser.addoption(
        OVERWRITE_SNAPSHOTS_CLI_FLAG,
        action="store_true",
        help="Overwrites existing snapshots by ones generated during this testing session",
    )


# In plan outputs, replace strings that vary from run to run with this so that comparisons can be made
# consistently.
PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS = "*"


def make_schema_replacement_function(system_schema: str, source_schema: str) -> Callable[[str], str]:
    """Generates a function to replace schema names in test outputs."""

    # The schema of the warehouse used in tests changes from run to run, so don't compare those.
    def replacement_function(text: str) -> str:
        # Replace with a string of the same length so that indents are preserved.
        text = text.replace(source_schema, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(source_schema))
        # Same with the MetricFlow system schema.
        return text.replace(system_schema, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(system_schema))

    return replacement_function


def replace_dataset_id_hash(text: str) -> str:
    """Replaces data set ID hashes for primed semantic models.

    The data set ID hash changes from run to run because it's based on the DW schema in the semantic model, which changes
    run to run.
    """
    pattern = re.compile(r"'[a-zA-Z0-9_]+__[a-zA-Z0-9_]+__(?P<hash>[a-zA-Z0-9_]+)'")
    while True:
        match = pattern.search(text)
        if match:
            data_set_id_hash = match.group("hash")
            # Replace with the same length to preserve indents
            text = text.replace(data_set_id_hash, PLACEHOLDER_CHAR_FOR_INCOMPARABLE_STRINGS * len(data_set_id_hash))
        else:
            break
    return text


PlanT = TypeVar("PlanT", bound=MetricFlowDag)


def assert_plan_snapshot_text_equal(
    request: FixtureRequest,
    mf_test_configuration: SnapshotConfiguration,
    plan: PlanT,
    plan_snapshot_text: str,
    plan_snapshot_file_extension: str = ".xml",
    exclude_line_regex: Optional[str] = None,
    incomparable_strings_replacement_function: Optional[Callable[[str], str]] = None,
    additional_sub_directories_for_snapshots: Tuple[str, ...] = (),
    expectation_description: Optional[str] = None,
) -> None:
    """Checks if the given plan text is equal to the one that's saved for comparison.

    * The location of the file is automatically generated based on the test and the plan's ID.
    * This may create a new saved plan file or overwrite the existing one, depending on the configuration.
    * replace_incomparable_strings is used to replace strings in the plan text before comparison. Useful for making
      plans consistent when there are strings that vary between runs and shouldn't be compared.
    * additional_sub_directories_for_snapshots is used to specify additional sub-directories (in the automatically
      generated directory) where plan outputs should reside.

    TODO: Make this more generic by renaming plan -> DAG.
    """
    assert_snapshot_text_equal(
        request=request,
        mf_test_configuration=mf_test_configuration,
        group_id=plan.__class__.__name__,
        snapshot_id=str(plan.dag_id),
        snapshot_text=plan_snapshot_text,
        snapshot_file_extension=plan_snapshot_file_extension,
        exclude_line_regex=exclude_line_regex,
        incomparable_strings_replacement_function=incomparable_strings_replacement_function,
        additional_sub_directories_for_snapshots=additional_sub_directories_for_snapshots,
        expectation_description=expectation_description,
    )


def assert_linkable_element_set_snapshot_equal(  # noqa: D103
    request: FixtureRequest,
    mf_test_configuration: SnapshotConfiguration,
    set_id: str,
    linkable_element_set: LinkableElementSet,
    expectation_description: Optional[str] = None,
) -> None:
    headers = ("Model Join-Path", "Entity Links", "Name", "Time Granularity", "Date Part", "Properties")
    rows = []
    for linkable_dimension_iterable in linkable_element_set.path_key_to_linkable_dimensions.values():
        for linkable_dimension in linkable_dimension_iterable:
            row_to_add = (
                # Checking a limited set of fields as the result is large due to the paths in the object.
                (linkable_dimension.join_path.left_semantic_model_reference.semantic_model_name,)
                + tuple(
                    path_element.semantic_model_reference.semantic_model_name
                    for path_element in linkable_dimension.join_path.path_elements
                ),
                tuple(entity_link.element_name for entity_link in linkable_dimension.entity_links),
                linkable_dimension.element_name,
                linkable_dimension.time_granularity.name if linkable_dimension.time_granularity is not None else "",
                linkable_dimension.date_part.name if linkable_dimension.date_part is not None else "",
                sorted(linkable_element_property.name for linkable_element_property in linkable_dimension.properties),
            )
            if row_to_add not in rows:
                rows.append(row_to_add)

    for linkable_entity_iterable in linkable_element_set.path_key_to_linkable_entities.values():
        for linkable_entity in linkable_entity_iterable:
            row_to_add = (
                # Checking a limited set of fields as the result is large due to the paths in the object.
                (linkable_entity.join_path.left_semantic_model_reference.semantic_model_name,)
                + tuple(
                    path_element.semantic_model_reference.semantic_model_name
                    for path_element in linkable_entity.join_path.path_elements
                ),
                tuple(entity_link.element_name for entity_link in linkable_entity.entity_links),
                linkable_entity.element_name,
                "",
                "",
                sorted(linkable_element_property.name for linkable_element_property in linkable_entity.properties),
            )
            if row_to_add not in rows:
                rows.append(row_to_add)

    for linkable_metric_iterable in linkable_element_set.path_key_to_linkable_metrics.values():
        for linkable_metric in linkable_metric_iterable:
            semantic_model_join_path = linkable_metric.join_path.semantic_model_join_path
            rows.append(
                (
                    # Checking a limited set of fields as the result is large due to the paths in the object.
                    (semantic_model_join_path.left_semantic_model_reference.semantic_model_name,)
                    + tuple(
                        path_element.semantic_model_reference.semantic_model_name
                        for path_element in semantic_model_join_path.path_elements
                    ),
                    (
                        str(tuple(entity_link.element_name for entity_link in linkable_metric.join_path.entity_links)),
                        str(
                            tuple(
                                entity_link.element_name for entity_link in linkable_metric.metric_subquery_entity_links
                            )
                        ),
                    ),
                    linkable_metric.element_name,
                    "",
                    "",
                    sorted(linkable_element_property.name for linkable_element_property in linkable_metric.properties),
                )
            )

    assert_str_snapshot_equal(
        request=request,
        mf_test_configuration=mf_test_configuration,
        snapshot_id=set_id,
        snapshot_str=tabulate.tabulate(headers=headers, tabular_data=sorted(rows)),
        expectation_description=expectation_description,
    )


def assert_spec_set_snapshot_equal(  # noqa: D103
    request: FixtureRequest,
    mf_test_configuration: SnapshotConfiguration,
    set_id: str,
    spec_set: InstanceSpecSet,
    expectation_description: Optional[str] = None,
) -> None:
    assert_object_snapshot_equal(
        request=request,
        mf_test_configuration=mf_test_configuration,
        obj_id=set_id,
        obj=sorted(spec.qualified_name for spec in spec_set.all_specs),
        expectation_description=expectation_description,
    )


def assert_linkable_spec_set_snapshot_equal(  # noqa: D103
    request: FixtureRequest,
    mf_test_configuration: SnapshotConfiguration,
    set_id: str,
    spec_set: LinkableSpecSet,
    expectation_description: Optional[str] = None,
) -> None:
    naming_scheme = ObjectBuilderNamingScheme()
    assert_snapshot_text_equal(
        request=request,
        mf_test_configuration=mf_test_configuration,
        group_id=spec_set.__class__.__name__,
        snapshot_id=set_id,
        snapshot_text=mf_pformat(sorted(naming_scheme.input_str(spec) for spec in spec_set.as_tuple)),
        snapshot_file_extension=".txt",
        additional_sub_directories_for_snapshots=(),
        expectation_description=expectation_description,
    )


def assert_object_snapshot_equal(  # type: ignore[misc]
    request: FixtureRequest,
    mf_test_configuration: SnapshotConfiguration,
    obj: Any,
    obj_id: str = "result",
    expectation_description: Optional[str] = None,
) -> None:
    """For tests to compare large objects, this can be used to snapshot a text representation of the object."""
    assert_snapshot_text_equal(
        request=request,
        mf_test_configuration=mf_test_configuration,
        group_id=obj.__class__.__name__,
        snapshot_id=obj_id,
        snapshot_text=mf_pformat(obj),
        snapshot_file_extension=".txt",
        expectation_description=expectation_description,
    )


def assert_str_snapshot_equal(  # noqa: D103
    request: FixtureRequest,
    mf_test_configuration: SnapshotConfiguration,
    snapshot_id: str,
    snapshot_str: str,
    expectation_description: Optional[str] = None,
    incomparable_strings_replacement_function: Optional[Callable[[str], str]] = None,
) -> None:
    """Write / compare a string snapshot."""
    assert_snapshot_text_equal(
        request=request,
        mf_test_configuration=mf_test_configuration,
        group_id=snapshot_str.__class__.__name__,
        snapshot_id=snapshot_id,
        snapshot_text=snapshot_str,
        snapshot_file_extension=".txt",
        expectation_description=expectation_description,
        incomparable_strings_replacement_function=incomparable_strings_replacement_function,
    )



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Dict, FrozenSet, Mapping, Sequence, Set

from typing_extensions import override

from metricflow.dataflow.dataflow_plan import DataflowPlan, DataflowPlanNode
from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitorWithDefaultHandler

logger = logging.getLogger(__name__)


class DataflowPlanAnalyzer:
    """Class to determine more complex properties of the dataflow plan.

    These could also be made as member methods of the dataflow plan, but this requires resolving some circular
    dependency issues to break out the functionality into separate files.
    """

    @staticmethod
    def find_common_branches(dataflow_plan: DataflowPlan) -> Sequence[DataflowPlanNode]:
        """Starting from the sink node, find the common branches that exist in the associated DAG.

        Returns a sorted sequence for reproducibility.
        """
        counting_visitor = _CountDataflowNodeVisitor()
        dataflow_plan.sink_node.accept(counting_visitor)

        node_to_common_count = counting_visitor.get_node_counts()

        common_nodes = []
        for node, count in node_to_common_count.items():
            if count > 1:
                common_nodes.append(node)

        common_branches_visitor = _FindLargestCommonBranchesVisitor(frozenset(common_nodes))

        return tuple(sorted(dataflow_plan.sink_node.accept(common_branches_visitor)))


class _CountDataflowNodeVisitor(DataflowPlanNodeVisitorWithDefaultHandler[None]):
    """Helper visitor to build a dict from a node in the plan to the number of times it appears in the plan."""

    def __init__(self) -> None:
        self._node_to_count: Dict[DataflowPlanNode, int] = defaultdict(int)

    def get_node_counts(self) -> Mapping[DataflowPlanNode, int]:
        return self._node_to_count

    @override
    def _default_handler(self, node: DataflowPlanNode) -> None:
        for parent_node in node.parent_nodes:
            parent_node.accept(self)
        self._node_to_count[node] += 1


class _FindLargestCommonBranchesVisitor(DataflowPlanNodeVisitorWithDefaultHandler[FrozenSet[DataflowPlanNode]]):
    """Given the nodes that are known to appear in the DAG multiple times, find the common branches.

    To get the largest common branches, (e.g. for `A -> B -> C -> D` and `B -> C -> D`, both `B -> C -> D`
    and `C -> D` can be considered common branches, and we want the largest one), this uses preorder traversal and
    returns the first common node that is seen.
    """

    def __init__(self, common_nodes: FrozenSet[DataflowPlanNode]) -> None:
        self._common_nodes = common_nodes

    @override
    def _default_handler(self, node: DataflowPlanNode) -> FrozenSet[DataflowPlanNode]:
        # Traversal starts from the leaf node and then goes to the parent branches. By doing this check first, we don't
        # return smaller common branches that are a part of a larger common branch.
        if node in self._common_nodes:
            return frozenset({node})

        common_branch_leaf_nodes: Set[DataflowPlanNode] = set()

        for parent_node in node.parent_nodes:
            common_branch_leaf_nodes.update(parent_node.accept(self))

        return frozenset(common_branch_leaf_nodes)



from __future__ import annotations

import logging
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SqlRequestId:
    """Identifies a request (i.e. a call to SqlClient.query() or SqlClient.execute()) to the SQL engine."""

    id_str: str

    def __repr__(self) -> str:  # noqa: D105
        return self.id_str



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional


class MetricFlowPrettyFormattable(ABC):
    """Changes behavior for pretty-formatting using `MetricFlowPrettyFormatter`.

    This interface is pending updates to allow for additional configuration and structured return types.
    """

    @property
    @abstractmethod
    def pretty_format(self) -> Optional[str]:
        """Return the pretty-formatted version of this object, or None if the default approach should be used."""
        raise NotImplementedError



"""Classes for modeling a DAG."""

from __future__ import annotations

import html
import logging
import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, Optional, Sequence, Tuple, TypeVar

import jinja2
from typing_extensions import override

from metricflow_semantics.dag.dag_to_text import MetricFlowDagTextFormatter
from metricflow_semantics.dag.id_prefix import IdPrefix
from metricflow_semantics.dag.sequential_id import SequentialIdGenerator
from metricflow_semantics.mf_logging.pretty_formattable import MetricFlowPrettyFormattable
from metricflow_semantics.visitor import VisitorOutputT

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class DisplayedProperty:  # type: ignore
    """When visualizing a node as text or a graphic, the properties to display with it.

    This should be displayed in the form "{key} = {str(value)}"
    """

    key: str
    value: Any  # type: ignore


@dataclass(frozen=True, order=True)
class NodeId:
    """Unique identifier for nodes in DAGs."""

    id_str: str

    def __repr__(self) -> str:  # noqa: D105
        return self.id_str

    @staticmethod
    def create_unique(id_prefix: IdPrefix) -> NodeId:  # noqa: D102
        return NodeId(str(SequentialIdGenerator.create_next_id(id_prefix)))


class DagNodeVisitor(Generic[VisitorOutputT], ABC):
    """An object that can be used to visit the nodes of a DAG graph."""

    @abstractmethod
    def visit_node(self, node: DagNode) -> VisitorOutputT:  # noqa: D102
        pass


DagNodeT = TypeVar("DagNodeT", bound="DagNode")


@dataclass(frozen=True, eq=False)
class DagNode(MetricFlowPrettyFormattable, Generic[DagNodeT], ABC):
    """A node in a DAG. These should be immutable.

    Since there should only be a single instance of a node with a given ID, `eq` can be set to false so that equality
    operations can be done without comparing the fields. Comparing the fields can be a slow process since the
    `parent_nodes` field is recursive.
    """

    parent_nodes: Tuple[DagNodeT, ...]

    def __post_init__(self) -> None:  # noqa: D105
        object.__setattr__(self, "_post_init_node_id", self.create_unique_id())

    @property
    def node_id(self) -> NodeId:
        """ID for uniquely identifying a given node.

        Ideally, this field would have a default value. However, setting a default field in this class means that all
        subclasses would have to have default values for all the fields as default fields must come at the end.
        This issue is resolved in Python 3.10 with `kw_only`, so this can be updated once this project's minimum Python
        version is 3.10.

        Set via `__setattr___` in  `__post__init__` to workaround limitations of frozen dataclasses.
        """
        return getattr(self, "_post_init_node_id")

    @property
    @abstractmethod
    def description(self) -> str:
        """A human-readable description for this node."""
        pass

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:
        """When the node is represented in a visualization, the properties to display with it."""
        return (
            DisplayedProperty("description", self.description),
            DisplayedProperty("node_id", self.node_id),
        )

    @property
    def graphviz_label(self) -> str:
        """When using graphviz to render this node, the label that should be used in the construction."""
        return make_graphviz_label(
            title=self.__class__.__name__,
            properties=self.displayed_properties,
        )

    @classmethod
    @abstractmethod
    def id_prefix(cls) -> IdPrefix:
        """The prefix to use when generating IDs for nodes.

        e.g. a prefix of "my_node" will generate an ID like "my_node_0"
        """
        pass

    @classmethod
    def create_unique_id(cls) -> NodeId:
        """Create and return a unique identifier to use when creating nodes."""
        return NodeId(id_str=SequentialIdGenerator.create_next_id(cls.id_prefix()).str_value)

    def accept_dag_node_visitor(self, visitor: DagNodeVisitor[VisitorOutputT]) -> VisitorOutputT:
        """Visit this node."""
        return visitor.visit_node(self)

    def structure_text(self, formatter: MetricFlowDagTextFormatter = MetricFlowDagTextFormatter()) -> str:
        """Return a text representation that shows the structure of the DAG component starting from this node."""
        return formatter.dag_component_to_text(self)

    @property
    @override
    def pretty_format(self) -> Optional[str]:
        return f"{self.__class__.__name__}(node_id={self.node_id.id_str})"


def make_graphviz_label(
    title: str, properties: Sequence[DisplayedProperty], title_font_size: int = 12, property_font_size: int = 6
) -> str:
    """Make a graphviz label that can be used for rendering to an image.

    The title will be in a large font, while the properties will be listed in a table in a smaller font.
    """
    # Convert all properties values into a HTML-safe string, then break the string into lines of 40 columns so that
    # the node boxes don't get so wide. Better to pretty-print the object, but unclear how to do so.
    formatted_properties = []
    for displayed_property in properties:
        lines = [html.escape(x) for x in textwrap.wrap(str(displayed_property.value), width=40)]
        formatted_properties.append(DisplayedProperty(displayed_property.key, "<BR/>".join(lines)))

    return jinja2.Template(
        # Formatting here: https://graphviz.org/doc/info/shapes.html#html
        textwrap.dedent(
            """\
            <<TABLE BORDER="0" CELLPADDING="1" CELLSPACING="0">
             <TR>
               <TD ALIGN="LEFT" BALIGN="LEFT" VALIGN="TOP" COLSPAN="2"><FONT point-size="{{ title_size }}">{{ title }}</FONT></TD>
             </TR>
             {%- for key, value in properties %}
             <TR>
               <TD ALIGN="LEFT" BALIGN="LEFT" VALIGN="TOP"><FONT point-size="{{ property_size }}">{{ key }}</FONT></TD>
               <TD ALIGN="LEFT" BALIGN="LEFT" VALIGN="TOP"><FONT point-size="{{ property_size }}">{{ value }}</FONT></TD>
             </TR>
             {%- endfor %}
            </TABLE>>
            """
        ),
        undefined=jinja2.StrictUndefined,
    ).render(
        title=title,
        title_size=title_font_size,
        property_size=property_font_size,
        properties=[(displayed_property.key, displayed_property.value) for displayed_property in formatted_properties],
    )


@dataclass(frozen=True)
class DagId:
    """Unique identifier for DAGs."""

    id_str: str

    def __str__(self) -> str:  # noqa: D105
        return self.id_str

    @staticmethod
    def from_str(id_str: str) -> DagId:
        """Migration helper to create DAG IDs."""
        return DagId(id_str)

    @staticmethod
    def from_id_prefix(id_prefix: IdPrefix) -> DagId:  # noqa: D102
        return DagId(id_str=SequentialIdGenerator.create_next_id(id_prefix).str_value)


class MetricFlowDag(Generic[DagNodeT]):
    """Represents a directed acyclic graph. The sink nodes will have the connected components."""

    def __init__(self, dag_id: DagId, sink_nodes: Sequence[DagNodeT]):  # noqa: D107
        self._dag_id = dag_id
        self._sink_nodes = tuple(sink_nodes)

    @property
    def dag_id(self) -> DagId:  # noqa: D102
        return self._dag_id

    @property
    def sink_nodes(self) -> Sequence[DagNodeT]:  # noqa: D102
        return self._sink_nodes

    def structure_text(self, formatter: MetricFlowDagTextFormatter = MetricFlowDagTextFormatter()) -> str:
        """Return a text representation that shows the structure of this DAG."""
        return formatter.dag_to_text(self)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple

from dbt_semantic_interfaces.dataclass_serialization import SerializableDataclass
from dbt_semantic_interfaces.protocols import WhereFilterIntersection

from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.query.group_by_item.filter_spec_resolution.filter_spec_lookup import (
    FilterSpecResolutionLookUp,
)
from metricflow_semantics.specs.dimension_spec import DimensionSpec
from metricflow_semantics.specs.entity_spec import EntitySpec
from metricflow_semantics.specs.group_by_metric_spec import GroupByMetricSpec
from metricflow_semantics.specs.linkable_spec_set import LinkableSpecSet
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.order_by_spec import OrderBySpec
from metricflow_semantics.specs.time_dimension_spec import TimeDimensionSpec


@dataclass(frozen=True)
class MetricFlowQuerySpec(SerializableDataclass):
    """Specs needed for running a query."""

    metric_specs: Tuple[MetricSpec, ...] = ()
    dimension_specs: Tuple[DimensionSpec, ...] = ()
    entity_specs: Tuple[EntitySpec, ...] = ()
    time_dimension_specs: Tuple[TimeDimensionSpec, ...] = ()
    group_by_metric_specs: Tuple[GroupByMetricSpec, ...] = ()
    order_by_specs: Tuple[OrderBySpec, ...] = ()
    time_range_constraint: Optional[TimeRangeConstraint] = None
    limit: Optional[int] = None
    filter_intersection: Optional[WhereFilterIntersection] = None
    filter_spec_resolution_lookup: FilterSpecResolutionLookUp = FilterSpecResolutionLookUp.empty_instance()
    min_max_only: bool = False

    @property
    def linkable_specs(self) -> LinkableSpecSet:  # noqa: D102
        return LinkableSpecSet(
            dimension_specs=self.dimension_specs,
            time_dimension_specs=self.time_dimension_specs,
            entity_specs=self.entity_specs,
            group_by_metric_specs=self.group_by_metric_specs,
        )

    def with_time_range_constraint(self, time_range_constraint: Optional[TimeRangeConstraint]) -> MetricFlowQuerySpec:
        """Return a query spec that's the same as self but with a different time_range_constraint."""
        return MetricFlowQuerySpec(
            metric_specs=self.metric_specs,
            dimension_specs=self.dimension_specs,
            entity_specs=self.entity_specs,
            time_dimension_specs=self.time_dimension_specs,
            group_by_metric_specs=self.group_by_metric_specs,
            order_by_specs=self.order_by_specs,
            time_range_constraint=time_range_constraint,
            limit=self.limit,
            filter_intersection=self.filter_intersection,
            filter_spec_resolution_lookup=self.filter_spec_resolution_lookup,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from metricflow_semantics.dag.id_prefix import IdPrefix, StaticIdPrefix
from metricflow_semantics.dag.mf_dag import DisplayedProperty
from metricflow_semantics.filters.time_constraint import TimeRangeConstraint
from metricflow_semantics.visitor import VisitorOutputT

from metricflow.dataflow.dataflow_plan_visitor import DataflowPlanNodeVisitor
from metricflow.dataflow.nodes.aggregate_measures import DataflowPlanNode


@dataclass(frozen=True, eq=False)
class ConstrainTimeRangeNode(DataflowPlanNode):
    """Constrains the time range of the input data set.

    For example, if the input data set had "sales by date", then this would restrict the data set so that it only
    includes sales for a specific range of dates.
    """

    time_range_constraint: TimeRangeConstraint

    def __post_init__(self) -> None:  # noqa: D105
        super().__post_init__()
        assert len(self.parent_nodes) == 1

    @staticmethod
    def create(  # noqa: D102
        parent_node: DataflowPlanNode,
        time_range_constraint: TimeRangeConstraint,
    ) -> ConstrainTimeRangeNode:
        return ConstrainTimeRangeNode(
            parent_nodes=(parent_node,),
            time_range_constraint=time_range_constraint,
        )

    @classmethod
    def id_prefix(cls) -> IdPrefix:  # noqa: D102
        return StaticIdPrefix.DATAFLOW_NODE_CONSTRAIN_TIME_RANGE_ID_PREFIX

    def accept(self, visitor: DataflowPlanNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_constrain_time_range_node(self)

    @property
    def description(self) -> str:  # noqa: D102
        return (
            f"Constrain Time Range to [{self.time_range_constraint.start_time.isoformat()}, "
            f"{self.time_range_constraint.end_time.isoformat()}]"
        )

    @property
    def parent_node(self) -> DataflowPlanNode:  # noqa: D102
        return self.parent_nodes[0]

    @property
    def displayed_properties(self) -> Sequence[DisplayedProperty]:  # noqa: D102
        return tuple(super().displayed_properties) + (
            DisplayedProperty("time_range_start", self.time_range_constraint.start_time.isoformat()),
            DisplayedProperty("time_range_end", self.time_range_constraint.end_time.isoformat()),
        )

    def functionally_identical(self, other_node: DataflowPlanNode) -> bool:  # noqa: D102
        return isinstance(other_node, self.__class__) and self.time_range_constraint == other_node.time_range_constraint

    def with_new_parents(self, new_parent_nodes: Sequence[DataflowPlanNode]) -> ConstrainTimeRangeNode:  # noqa: D102
        assert len(new_parent_nodes) == 1
        return ConstrainTimeRangeNode(
            parent_nodes=tuple(new_parent_nodes),
            time_range_constraint=self.time_range_constraint,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from metricflow_semantics.sql.sql_table import SqlTable


@dataclass(frozen=True, order=True)
class SqlColumn:
    """Represents a reference to a SQL column."""

    table: SqlTable
    column_name: str

    @staticmethod
    def from_names(
        column_name: str,
        table_name: str,
        schema_name: str,
        db_name: str,
    ) -> SqlColumn:
        """Helper factory method for constructing a column from database, table, schema and column names."""
        table = SqlTable(db_name=db_name, schema_name=schema_name, table_name=table_name)
        return SqlColumn(table=table, column_name=column_name)

    @staticmethod
    def from_string(sql_str: str) -> SqlColumn:  # noqa: D102
        table_str, column_name = sql_str.rsplit(".", 1)
        table = SqlTable.from_string(table_str)
        return SqlColumn(table=table, column_name=column_name)

    @property
    def db_name(self) -> Optional[str]:  # noqa: D102
        return self.table.db_name

    @property
    def schema_name(self) -> Optional[str]:  # noqa: D102
        return self.table.schema_name

    @property
    def table_name(self) -> str:  # noqa: D102
        return self.table.table_name

    @property
    def sql(self) -> str:
        """Return the snippet that can be used for use in SQL queries."""
        return f"{self.table.sql}.{self.column_name}"

    def __repr__(self) -> str:  # noqa: D105
        return f"SqlColumn(full_name={self.sql})"



from __future__ import annotations


def assert_exactly_one_arg_set(**kwargs) -> None:  # type: ignore
    """Throws an assertion error if 0 or more than 1 argument is not None."""
    num_set = 0
    for value in kwargs.values():
        if value is not None:
            num_set += 1

    assert num_set == 1, f"{num_set} argument(s) set instead of 1 in arguments: {kwargs}"


def assert_at_most_one_arg_set(**kwargs) -> None:  # type: ignore
    """Throws an assertion error if more than 1 argument is not None."""
    num_set = 0
    for value in kwargs.values():
        if value is not None:
            num_set += 1

    assert num_set <= 1, f"{num_set} argument(s) set instead of <=1 in arguments: {kwargs}"



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.query.group_by_item.path_prefixable import PathPrefixable
from metricflow_semantics.query.group_by_item.resolution_dag.resolution_nodes.base_node import GroupByItemResolutionNode


@dataclass(frozen=True)
class MetricFlowQueryResolutionPath(PathPrefixable):
    """Represents a path in the group-by-item resolution DAG."""

    resolution_path_nodes: Tuple[GroupByItemResolutionNode, ...]

    @staticmethod
    def empty_instance() -> MetricFlowQueryResolutionPath:  # noqa: D102
        return MetricFlowQueryResolutionPath(
            resolution_path_nodes=(),
        )

    @property
    def last_item(self) -> GroupByItemResolutionNode:
        """Return the last node in this path."""
        return self.resolution_path_nodes[-1]

    @property
    def ui_description(self) -> str:  # noqa: D102
        if len(self.resolution_path_nodes) == 0:
            return "[Empty Path]"
        descriptions = tuple(f"[Resolve {path_node.ui_description}]" for path_node in self.resolution_path_nodes)
        output = descriptions[0]

        for i, description in enumerate(descriptions[1:]):
            output += "\n"
            output += indent("-> " + description, indent_level=i + 1)

        return output

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> MetricFlowQueryResolutionPath:
        return MetricFlowQueryResolutionPath(
            resolution_path_nodes=path_prefix.resolution_path_nodes + self.resolution_path_nodes
        )

    @override
    def __str__(self) -> str:
        items = [self.__class__.__name__, "(", ", ".join(tuple(str(node) for node in self.resolution_path_nodes)), ")"]
        return "".join(items)

    @staticmethod
    def from_path_item(node: GroupByItemResolutionNode) -> MetricFlowQueryResolutionPath:
        """Creates a single element path with the given node."""
        return MetricFlowQueryResolutionPath(
            resolution_path_nodes=(node,),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from dbt_semantic_interfaces.implementations.metric import PydanticMetricTimeWindow
from dbt_semantic_interfaces.references import MetricReference
from dbt_semantic_interfaces.type_enums import TimeGranularity

from metricflow_semantics.specs.instance_spec import InstanceSpec, InstanceSpecVisitor
from metricflow_semantics.specs.where_filter.where_filter_spec_set import WhereFilterSpecSet
from metricflow_semantics.visitor import VisitorOutputT


@dataclass(frozen=True)
class MetricSpec(InstanceSpec):  # noqa: D101
    # Time-over-time could go here
    element_name: str
    filter_spec_set: WhereFilterSpecSet = WhereFilterSpecSet()
    alias: Optional[str] = None
    offset_window: Optional[PydanticMetricTimeWindow] = None
    offset_to_grain: Optional[TimeGranularity] = None

    @staticmethod
    def from_element_name(element_name: str) -> MetricSpec:  # noqa: D102
        return MetricSpec(element_name=element_name)

    @property
    def qualified_name(self) -> str:  # noqa: D102
        return self.element_name

    @staticmethod
    def from_reference(reference: MetricReference) -> MetricSpec:
        """Initialize from a metric reference instance."""
        return MetricSpec(element_name=reference.element_name)

    def accept(self, visitor: InstanceSpecVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D102
        return visitor.visit_metric_spec(self)

    @property
    def reference(self) -> MetricReference:
        """Return the reference object that is used for referencing the associated metric in the manifest."""
        return MetricReference(element_name=self.element_name)

    @property
    def has_time_offset(self) -> bool:  # noqa: D102
        return bool(self.offset_window or self.offset_to_grain)

    def without_offset(self) -> MetricSpec:
        """Represents the metric spec with any time offsets removed."""
        return MetricSpec(element_name=self.element_name, filter_spec_set=self.filter_spec_set, alias=self.alias)

    def with_alias(self, alias: Optional[str]) -> MetricSpec:
        """Add the alias to the metric spec."""
        return MetricSpec(
            element_name=self.element_name,
            filter_spec_set=self.filter_spec_set,
            alias=alias,
            offset_window=self.offset_window,
            offset_to_grain=self.offset_to_grain,
        )

    def without_filter_specs(self) -> MetricSpec:  # noqa: D102
        return MetricSpec(
            element_name=self.element_name,
            alias=self.alias,
            offset_window=self.offset_window,
            offset_to_grain=self.offset_to_grain,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from typing_extensions import override

from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.mf_logging.pretty_print import mf_pformat
from metricflow_semantics.naming.object_builder_scheme import ObjectBuilderNamingScheme
from metricflow_semantics.query.group_by_item.candidate_push_down.group_by_item_candidate import GroupByItemCandidateSet
from metricflow_semantics.query.group_by_item.resolution_path import MetricFlowQueryResolutionPath
from metricflow_semantics.query.issues.issues_base import (
    MetricFlowQueryIssueType,
    MetricFlowQueryResolutionIssue,
)
from metricflow_semantics.query.resolver_inputs.base_resolver_inputs import MetricFlowQueryResolverInput


@dataclass(frozen=True)
class MultipleMatchIssue(MetricFlowQueryResolutionIssue):
    """Describes an issue during group-by-item resolution the input pattern matches multiple specs."""

    candidate_set: GroupByItemCandidateSet

    @staticmethod
    def from_parameters(  # noqa: D102
        query_resolution_path: MetricFlowQueryResolutionPath,
        candidate_set: GroupByItemCandidateSet,
        parent_issues: Sequence[MetricFlowQueryResolutionIssue],
    ) -> MultipleMatchIssue:
        return MultipleMatchIssue(
            issue_type=MetricFlowQueryIssueType.ERROR,
            query_resolution_path=query_resolution_path,
            candidate_set=candidate_set,
            parent_issues=tuple(parent_issues),
        )

    @override
    def ui_description(self, associated_input: MetricFlowQueryResolverInput) -> str:
        last_path_item = self.query_resolution_path.last_item
        naming_scheme = (
            associated_input.input_pattern_description.naming_scheme
            if associated_input.input_pattern_description is not None
            else ObjectBuilderNamingScheme()
        )

        specs_as_strs = []
        for spec in self.candidate_set.specs:
            input_str = naming_scheme.input_str(spec)
            if input_str is not None:
                specs_as_strs.append(input_str)
            else:
                specs_as_strs.append(f"<{repr(spec)}>")

        return (
            f"The given input matches multiple group-by-items for {last_path_item.ui_description}:\n\n"
            f"{indent(mf_pformat(specs_as_strs))}\n\n"
            f"Please use a more specific input to resolve the ambiguity."
        )

    @override
    def with_path_prefix(self, path_prefix: MetricFlowQueryResolutionPath) -> MultipleMatchIssue:
        return MultipleMatchIssue(
            issue_type=self.issue_type,
            parent_issues=tuple(issue.with_path_prefix(path_prefix) for issue in self.parent_issues),
            query_resolution_path=self.query_resolution_path.with_path_prefix(path_prefix),
            candidate_set=self.candidate_set.with_path_prefix(path_prefix),
        )



from __future__ import annotations

import logging
from contextlib import contextmanager
from typing import Iterator

from metricflow_semantics.mf_logging.lazy_formattable import LazyFormat
from typing_extensions import override

from metricflow.sql.optimizer.cte_alias_to_cte_node_mapping import SqlCteAliasMappingLookup
from metricflow.sql.sql_plan import (
    SqlCreateTableAsNode,
    SqlCteAliasMapping,
    SqlCteNode,
    SqlPlanNode,
    SqlPlanNodeVisitor,
    SqlSelectQueryFromClauseNode,
    SqlSelectStatementNode,
    SqlTableNode,
)

logger = logging.getLogger(__name__)


class SqlCteAliasMappingLookupBuilderVisitor(SqlPlanNodeVisitor[None]):
    """Traverses the SQL plan and builds the associated `SqlCteAliasMappingLookup`.

    Please see `SqlCteAliasMappingLookup` for more details.
    """

    def __init__(self) -> None:  # noqa: D107
        self._current_cte_alias_mapping = SqlCteAliasMapping()
        self._cte_alias_mapping_lookup = SqlCteAliasMappingLookup()

    @contextmanager
    def _save_current_cte_alias_mapping(self) -> Iterator[None]:
        previous_cte_alias_mapping = self._current_cte_alias_mapping
        yield
        self._current_cte_alias_mapping = previous_cte_alias_mapping

    def _default_handler(self, node: SqlPlanNode) -> None:
        """Default recursive handler to visit the parents of the given node."""
        for parent_node in node.parent_nodes:
            with self._save_current_cte_alias_mapping():
                parent_node.accept(self)
        return

    @override
    def visit_cte_node(self, node: SqlCteNode) -> None:
        return self._default_handler(node)

    @override
    def visit_select_statement_node(self, node: SqlSelectStatementNode) -> None:
        """Based on required column aliases for this SELECT, figure out required column aliases in parents."""
        logger.debug(
            LazyFormat(
                "Starting visit of SELECT statement node with CTE alias mapping",
                node=node,
                current_cte_alias_mapping=self._current_cte_alias_mapping,
            )
        )

        if self._cte_alias_mapping_lookup.cte_alias_mapping_exists(node):
            return self._default_handler(node)

        # Record that can see the CTEs defined in this SELECT node and outer SELECT statements.
        # CTEs defined in this select node should override ones that were defined in the outer SELECT in case
        # of CTE alias collisions.
        self._current_cte_alias_mapping = self._current_cte_alias_mapping.merge(
            SqlCteAliasMapping.create({cte_node.cte_alias: cte_node for cte_node in node.cte_sources})
        )
        self._cte_alias_mapping_lookup.add_cte_alias_mapping(
            select_node=node,
            cte_alias_mapping=self._current_cte_alias_mapping,
        )

        return self._default_handler(node)

    @override
    def visit_table_node(self, node: SqlTableNode) -> None:
        self._default_handler(node)

    @override
    def visit_query_from_clause_node(self, node: SqlSelectQueryFromClauseNode) -> None:
        self._default_handler(node)

    @override
    def visit_create_table_as_node(self, node: SqlCreateTableAsNode) -> None:  # noqa: D102
        self._default_handler(node)

    @property
    def cte_alias_mapping_lookup(self) -> SqlCteAliasMappingLookup:
        """Returns the lookup created after traversal."""
        return self._cte_alias_mapping_lookup



from __future__ import annotations

from dataclasses import dataclass
from typing import Sequence

from dbt_semantic_interfaces.references import MetricReference
from typing_extensions import override

from metricflow_semantics.specs.instance_spec import InstanceSpec
from metricflow_semantics.specs.metric_spec import MetricSpec
from metricflow_semantics.specs.patterns.spec_pattern import SpecPattern
from metricflow_semantics.specs.spec_set import group_specs_by_type


@dataclass(frozen=True)
class MetricSpecPattern(SpecPattern):
    """Matches MetricSpecs that have the given metric_reference."""

    metric_reference: MetricReference

    @override
    def match(self, candidate_specs: Sequence[InstanceSpec]) -> Sequence[MetricSpec]:
        spec_set = group_specs_by_type(candidate_specs)
        return tuple(
            metric_name for metric_name in spec_set.metric_specs if metric_name.reference == self.metric_reference
        )



from __future__ import annotations

import logging
import textwrap
from abc import ABC, abstractmethod
from collections import namedtuple
from dataclasses import dataclass
from typing import TYPE_CHECKING, Collection, List, Optional

import jinja2
from dbt_semantic_interfaces.type_enums.date_part import DatePart
from dbt_semantic_interfaces.type_enums.time_granularity import TimeGranularity
from metricflow_semantics.mf_logging.formatting import indent
from metricflow_semantics.sql.sql_bind_parameters import SqlBindParameterSet
from metricflow_semantics.sql.sql_exprs import (
    SqlAddTimeExpression,
    SqlAggregateFunctionExpression,
    SqlArithmeticExpression,
    SqlArithmeticOperator,
    SqlBetweenExpression,
    SqlCaseExpression,
    SqlCastToTimestampExpression,
    SqlColumnAliasReferenceExpression,
    SqlColumnReferenceExpression,
    SqlComparisonExpression,
    SqlDateTruncExpression,
    SqlExpressionNode,
    SqlExpressionNodeVisitor,
    SqlExtractExpression,
    SqlFunction,
    SqlGenerateUuidExpression,
    SqlIntegerExpression,
    SqlIsNullExpression,
    SqlLogicalExpression,
    SqlNullExpression,
    SqlPercentileExpression,
    SqlPercentileFunctionType,
    SqlRatioComputationExpression,
    SqlStringExpression,
    SqlStringLiteralExpression,
    SqlSubtractTimeIntervalExpression,
    SqlWindowFunctionExpression,
)
from typing_extensions import override

from metricflow.sql.render.rendering_constants import SqlRenderingConstants
from metricflow.sql.sql_plan import SqlSelectColumn

if TYPE_CHECKING:
    from metricflow.protocols.sql_client import SqlEngine


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SqlExpressionRenderResult:
    """The result of rendering an SQL expression tree to a string."""

    sql: str
    bind_parameter_set: SqlBindParameterSet


class SqlExpressionRenderer(SqlExpressionNodeVisitor[SqlExpressionRenderResult], ABC):
    """Renders SqlExpressions into strings."""

    def render_sql_expr(self, sql_expr: SqlExpressionNode) -> SqlExpressionRenderResult:
        """Render the given expression to a string."""
        return sql_expr.accept(self)

    def render_group_by_expr(self, group_by_column: SqlSelectColumn) -> SqlExpressionRenderResult:
        """Render the input group by column to a string.

        This allows for engine-level overrides of the group by rendering behavior, since some engines only support
        rendering group by columns based on aliases.
        """
        return self.render_sql_expr(sql_expr=group_by_column.expr)

    @property
    @abstractmethod
    def double_data_type(self) -> str:
        """Property for the double data type, for engine-specific type casting."""
        raise NotImplementedError

    @property
    @abstractmethod
    def timestamp_data_type(self) -> str:
        """Property for the timestamp data type, for engine-specific type casting."""
        raise NotImplementedError

    @property
    @abstractmethod
    def supported_percentile_function_types(self) -> Collection[SqlPercentileFunctionType]:
        """Returns the sequence of supported percentile function types for the implementing renderer."""
        raise NotImplementedError

    def can_render_percentile_function(self, percentile_type: SqlPercentileFunctionType) -> bool:
        """Whether or not this SQL renderer supports rendering the particular percentile function type."""
        return percentile_type in self.supported_percentile_function_types


class DefaultSqlExpressionRenderer(SqlExpressionRenderer):
    """Renders the SQL query plan assuming ANSI SQL."""

    @property
    def sql_engine(self) -> Optional[SqlEngine]:  # noqa: D102
        return None

    @property
    @override
    def double_data_type(self) -> str:
        return "DOUBLE"

    @property
    @override
    def timestamp_data_type(self) -> str:
        return "TIMESTAMP"

    @property
    @override
    def supported_percentile_function_types(self) -> Collection[SqlPercentileFunctionType]:
        return {}

    def visit_string_expr(self, node: SqlStringExpression) -> SqlExpressionRenderResult:
        """Renders an arbitrary string expression like 1+1=2."""
        return SqlExpressionRenderResult(sql=node.sql_expr, bind_parameter_set=node.bind_parameter_set)

    def visit_column_reference_expr(self, node: SqlColumnReferenceExpression) -> SqlExpressionRenderResult:
        """Render a reference to a column in a table like my_table.my_col."""
        return SqlExpressionRenderResult(
            sql=(
                f"{node.col_ref.table_alias}.{node.col_ref.column_name}"
                if node.should_render_table_alias
                else node.col_ref.column_name
            ),
            bind_parameter_set=SqlBindParameterSet(),
        )

    def visit_column_alias_reference_expr(self, node: SqlColumnAliasReferenceExpression) -> SqlExpressionRenderResult:
        """Render a reference to a column without a known table alias. e.g. foo.bar vs bar."""
        return SqlExpressionRenderResult(
            sql=node.column_alias,
            bind_parameter_set=SqlBindParameterSet(),
        )

    def visit_comparison_expr(self, node: SqlComparisonExpression) -> SqlExpressionRenderResult:
        """Render a comparison expression like 1 = 2."""
        combined_params = SqlBindParameterSet()

        left_expr_rendered = self.render_sql_expr(node.left_expr)
        combined_params = combined_params.merge(left_expr_rendered.bind_parameter_set)

        right_expr_rendered = self.render_sql_expr(node.right_expr)
        combined_params = combined_params.merge(right_expr_rendered.bind_parameter_set)

        # To avoid issues with operator precedence, use parenthesis to group the left / right expressions if they
        # contain operators.
        return SqlExpressionRenderResult(
            # Render a + b = c
            sql=(
                (f"({left_expr_rendered.sql})" if node.left_expr.requires_parenthesis else left_expr_rendered.sql)
                + f" {node.comparison.value} "
                + (f"({right_expr_rendered.sql})" if node.right_expr.requires_parenthesis else right_expr_rendered.sql)
            ),
            bind_parameter_set=combined_params,
        )

    def visit_function_expr(self, node: SqlAggregateFunctionExpression) -> SqlExpressionRenderResult:
        """Render a function call like CONCAT(a, b)."""
        args_rendered = [self.render_sql_expr(x) for x in node.sql_function_args]
        combined_params = SqlBindParameterSet()
        for arg_rendered in args_rendered:
            combined_params = combined_params.merge(arg_rendered.bind_parameter_set)

        distinct_prefix = "DISTINCT " if SqlFunction.is_distinct_aggregation(node.sql_function) else ""
        args_string = ", ".join([x.sql for x in args_rendered])

        return SqlExpressionRenderResult(
            sql=f"{node.sql_function.value}({distinct_prefix}{args_string})",
            bind_parameter_set=combined_params,
        )

    def visit_percentile_expr(self, node: SqlPercentileExpression) -> SqlExpressionRenderResult:
        """Render a percentile expression."""
        raise RuntimeError(
            "Default expression render has no percentile implementation - an engine-specific renderer should be implemented."
        )

    def visit_null_expr(self, node: SqlNullExpression) -> SqlExpressionRenderResult:  # noqa: D102
        return SqlExpressionRenderResult(
            sql="NULL",
            bind_parameter_set=SqlBindParameterSet(),
        )

    def visit_string_literal_expr(self, node: SqlStringLiteralExpression) -> SqlExpressionRenderResult:  # noqa: D102
        return SqlExpressionRenderResult(
            sql=f"'{node.literal_value}'",
            bind_parameter_set=SqlBindParameterSet(),
        )

    def visit_logical_expr(self, node: SqlLogicalExpression) -> SqlExpressionRenderResult:  # noqa: D102
        RenderedExpr = namedtuple("RenderedExpr", ["expr", "requires_parenthesis"])
        args_rendered = [RenderedExpr(self.render_sql_expr(x), x.requires_parenthesis) for x in node.args]
        combined_parameters = SqlBindParameterSet()
        args_sql: List[str] = []

        can_be_rendered_in_one_line = sum(len(x.expr.sql) for x in args_rendered) < 60

        for arg_rendered in args_rendered:
            combined_parameters.merge(arg_rendered.expr.bind_parameter_set)
            arg_sql = self._render_logical_arg(
                arg_rendered.expr, arg_rendered.requires_parenthesis, render_in_one_line=can_be_rendered_in_one_line
            )
            args_sql.append(arg_sql)

        sql = f" {node.operator.value} ".join(args_sql)

        return SqlExpressionRenderResult(
            sql=sql,
            bind_parameter_set=combined_parameters,
        )

    @staticmethod
    def _render_logical_arg(
        arg_rendered: SqlExpressionRenderResult, requires_parenthesis: bool, render_in_one_line: bool
    ) -> str:
        # Put everything on 1 line for short expressions, but otherwise put one expression per line
        # Note: multi-line expressions are always enclosed in parentheses.

        # e.g.
        # Put everything on 1 line for short expressions like:
        #
        # (1 < 2) AND foo
        #
        # but for long expressions do:
        #
        #  (
        #    some_long_expression1
        #  ) AND (
        #    some_long_expression2
        #  ) AND (
        #    some_long_expression3
        # )

        if render_in_one_line:
            return arg_rendered.sql if not requires_parenthesis else f"({arg_rendered.sql})"
        else:
            return (
                jinja2.Template(
                    textwrap.dedent(
                        """\
                    (
                      {{ arg_sql | indent(2) }}
                    )
                    """
                    )
                )
                .render(arg_sql=arg_rendered.sql)
                .rstrip()
            )

    def visit_is_null_expr(self, node: SqlIsNullExpression) -> SqlExpressionRenderResult:  # noqa: D102
        arg_rendered = self.render_sql_expr(node.arg)

        return SqlExpressionRenderResult(
            sql=f"{arg_rendered.sql} IS NULL" if not node.arg.requires_parenthesis else f"({arg_rendered.sql}) IS NULL",
            bind_parameter_set=arg_rendered.bind_parameter_set,
        )

    def visit_cast_to_timestamp_expr(  # noqa: D102
        self, node: SqlCastToTimestampExpression
    ) -> SqlExpressionRenderResult:  # noqa: D102
        arg_rendered = self.render_sql_expr(node.arg)
        return SqlExpressionRenderResult(
            sql=f"CAST({arg_rendered.sql} AS {self.timestamp_data_type})",
            bind_parameter_set=arg_rendered.bind_parameter_set,
        )

    def _validate_granularity_for_engine(self, time_granularity: TimeGranularity) -> None:
        if self.sql_engine and time_granularity in self.sql_engine.unsupported_granularities:
            raise RuntimeError(f"{self.sql_engine.name} does not support time granularity {time_granularity.name}.")

    def visit_date_trunc_expr(self, node: SqlDateTruncExpression) -> SqlExpressionRenderResult:  # noqa: D102
        self._validate_granularity_for_engine(node.time_granularity)

        arg_rendered = self.render_sql_expr(node.arg)

        return SqlExpressionRenderResult(
            sql=f"DATE_TRUNC('{node.time_granularity.value}', {arg_rendered.sql})",
            bind_parameter_set=arg_rendered.bind_parameter_set,
        )

    def visit_extract_expr(self, node: SqlExtractExpression) -> SqlExpressionRenderResult:  # noqa: D102
        arg_rendered = self.render_sql_expr(node.arg)

        return SqlExpressionRenderResult(
            sql=f"EXTRACT({self.render_date_part(node.date_part)} FROM {arg_rendered.sql})",
            bind_parameter_set=arg_rendered.bind_parameter_set,
        )

    def render_date_part(self, date_part: DatePart) -> str:
        """Render DATE PART for an EXTRACT expression.

        For DatePart.DOW (day of week) we use the ISO date part to ensure all engines return consistent results.
        """
        if date_part is DatePart.DOW:
            return "isodow"

        return date_part.value

    def visit_subtract_time_interval_expr(  # noqa: D102
        self, node: SqlSubtractTimeIntervalExpression
    ) -> SqlExpressionRenderResult:
        arg_rendered = node.arg.accept(self)

        count = node.count
        granularity = node.granularity
        if granularity is TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            count *= 3
        return SqlExpressionRenderResult(
            sql=f"DATEADD({granularity.value}, -{count}, {arg_rendered.sql})",
            bind_parameter_set=arg_rendered.bind_parameter_set,
        )

    def visit_add_time_expr(self, node: SqlAddTimeExpression) -> SqlExpressionRenderResult:  # noqa: D102
        granularity = node.granularity
        count_expr = node.count_expr
        if granularity is TimeGranularity.QUARTER:
            granularity = TimeGranularity.MONTH
            count_expr = SqlArithmeticExpression.create(
                left_expr=node.count_expr,
                operator=SqlArithmeticOperator.MULTIPLY,
                right_expr=SqlIntegerExpression.create(3),
            )

        arg_rendered = node.arg.accept(self)
        count_rendered = count_expr.accept(self)
        count_sql = f"({count_rendered.sql})" if count_expr.requires_parenthesis else count_rendered.sql

        return SqlExpressionRenderResult(
            sql=f"DATEADD({granularity.value}, {count_sql}, {arg_rendered.sql})",
            bind_parameter_set=SqlBindParameterSet.merge_iterable(
                (arg_rendered.bind_parameter_set, count_rendered.bind_parameter_set)
            ),
        )

    def visit_ratio_computation_expr(self, node: SqlRatioComputationExpression) -> SqlExpressionRenderResult:
        """Render the ratio computation for a ratio metric.

        This requires both a type cast to a floating point type (default to DOUBLE, engine-permitting) and
        the requisite division between numerator and denominator
        """
        rendered_numerator = self.render_sql_expr(node.numerator)
        rendered_denominator = self.render_sql_expr(node.denominator)

        numerator_sql = f"CAST({rendered_numerator.sql} AS {self.double_data_type})"
        denominator_sql = f"CAST(NULLIF({rendered_denominator.sql}, 0) AS {self.double_data_type})"

        bind_parameter_set = SqlBindParameterSet()
        bind_parameter_set = bind_parameter_set.merge(rendered_numerator.bind_parameter_set)
        bind_parameter_set = bind_parameter_set.merge(rendered_denominator.bind_parameter_set)

        return SqlExpressionRenderResult(
            sql=f"{numerator_sql} / {denominator_sql}",
            bind_parameter_set=bind_parameter_set,
        )

    def visit_between_expr(self, node: SqlBetweenExpression) -> SqlExpressionRenderResult:  # noqa: D102
        rendered_column_arg = self.render_sql_expr(node.column_arg)
        rendered_start_expr = self.render_sql_expr(node.start_expr)
        rendered_end_expr = self.render_sql_expr(node.end_expr)

        bind_parameter_set = SqlBindParameterSet()
        bind_parameter_set = bind_parameter_set.merge(rendered_column_arg.bind_parameter_set)
        bind_parameter_set = bind_parameter_set.merge(rendered_start_expr.bind_parameter_set)
        bind_parameter_set = bind_parameter_set.merge(rendered_end_expr.bind_parameter_set)

        return SqlExpressionRenderResult(
            sql=f"{rendered_column_arg.sql} BETWEEN {rendered_start_expr.sql} AND {rendered_end_expr.sql}",
            bind_parameter_set=bind_parameter_set,
        )

    def visit_window_function_expr(self, node: SqlWindowFunctionExpression) -> SqlExpressionRenderResult:  # noqa: D102
        sql_function_args_rendered = [self.render_sql_expr(x) for x in node.sql_function_args]
        partition_by_args_rendered = [self.render_sql_expr(x) for x in node.partition_by_args]
        order_by_args_rendered = {self.render_sql_expr(x.expr): x for x in node.order_by_args}

        combined_params = SqlBindParameterSet()
        args_rendered = []
        if sql_function_args_rendered:
            args_rendered.extend(sql_function_args_rendered)
        if partition_by_args_rendered:
            args_rendered.extend(partition_by_args_rendered)
        if order_by_args_rendered:
            args_rendered.extend(list(order_by_args_rendered.keys()))
        for arg_rendered in args_rendered:
            combined_params = combined_params.merge(arg_rendered.bind_parameter_set)

        sql_function_args_string = ", ".join([x.sql for x in sql_function_args_rendered])
        window_string_lines: List[str] = []
        if len(partition_by_args_rendered) == 1:
            window_string_lines.append(f"PARTITION BY {partition_by_args_rendered[0].sql}")
        elif len(partition_by_args_rendered) > 1:
            window_string_lines.append("PARTITION BY")
            window_string_lines.append(
                indent(
                    "\n, ".join([x.sql for x in partition_by_args_rendered]),
                    indent_prefix=SqlRenderingConstants.INDENT,
                )
            )
        if len(order_by_args_rendered) == 1:
            rendered_result, order_by_arg = tuple(order_by_args_rendered.items())[0]
            window_string_lines.append(
                "ORDER BY " + rendered_result.sql + (f" {order_by_arg.suffix}" if order_by_arg.suffix else "")
            )
        elif len(order_by_args_rendered) > 1:
            window_string_lines.append("ORDER BY")
            window_string_lines.append(
                indent(
                    "\n, ".join(
                        [
                            rendered_result.sql + (f" {order_by_arg.suffix}" if order_by_arg.suffix else "")
                            for rendered_result, order_by_arg in order_by_args_rendered.items()
                        ]
                    ),
                    indent_prefix=SqlRenderingConstants.INDENT,
                )
            )

        if len(order_by_args_rendered) > 0 and node.sql_function.allows_frame_clause:
            window_string_lines.append("ROWS BETWEEN UNBOUNDED PRECEDING AND UNBOUNDED FOLLOWING")

        window_string = "\n".join(window_string_lines)

        if len(window_string_lines) <= 1:
            return SqlExpressionRenderResult(
                sql=f"{node.sql_function.value}({sql_function_args_string}) OVER ({window_string})",
                bind_parameter_set=combined_params,
            )
        else:
            indented_window_string = indent(window_string, indent_prefix=SqlRenderingConstants.INDENT)
            return SqlExpressionRenderResult(
                sql=f"{node.sql_function.value}({sql_function_args_string}) OVER (\n{indented_window_string}\n)",
                bind_parameter_set=combined_params,
            )

    def visit_generate_uuid_expr(self, node: SqlGenerateUuidExpression) -> SqlExpressionRenderResult:  # noqa: D102
        return SqlExpressionRenderResult(
            sql="UUID()",
            bind_parameter_set=SqlBindParameterSet(),
        )

    def visit_case_expr(self, node: SqlCaseExpression) -> SqlExpressionRenderResult:  # noqa: D102
        sql = "CASE\n"
        for when, then in node.when_to_then_exprs.items():
            sql += indent(
                f"WHEN {self.render_sql_expr(when).sql}\n", indent_prefix=SqlRenderingConstants.INDENT
            ) + indent(
                f"THEN {self.render_sql_expr(then).sql}\n",
                indent_prefix=SqlRenderingConstants.INDENT * 2,
            )
        if node.else_expr:
            sql += indent(
                f"ELSE {self.render_sql_expr(node.else_expr).sql}\n",
                indent_prefix=SqlRenderingConstants.INDENT,
            )
        sql += "END"
        return SqlExpressionRenderResult(sql=sql, bind_parameter_set=SqlBindParameterSet())

    def visit_arithmetic_expr(self, node: SqlArithmeticExpression) -> SqlExpressionRenderResult:  # noqa: D102
        sql = f"{self.render_sql_expr(node.left_expr).sql} {node.operator.value} {self.render_sql_expr(node.right_expr).sql}"
        return SqlExpressionRenderResult(sql=sql, bind_parameter_set=SqlBindParameterSet())

    def visit_integer_expr(self, node: SqlIntegerExpression) -> SqlExpressionRenderResult:  # noqa: D102
        return SqlExpressionRenderResult(sql=str(node.integer_value), bind_parameter_set=SqlBindParameterSet())
